"""Instruments for the OpenTelemetry plugin."""

# pyright: reportMissingTypeStubs=false

from collections.abc import Callable
from importlib.util import find_spec
from typing import Any

from opentelemetry.sdk.metrics import MeterProvider
from opentelemetry.sdk.trace import TracerProvider

from fastapi_factory_utilities.core.plugins.opentelemetry_plugin.configs import OpenTelemetryConfig
from fastapi_factory_utilities.core.protocols import ApplicationAbstractProtocol


def instrument_fastapi(
    application: ApplicationAbstractProtocol,
    config: OpenTelemetryConfig,
    meter_provider: MeterProvider,
    tracer_provider: TracerProvider,
) -> None:
    """Instrument the FastAPI application."""
    if find_spec(name="fastapi") and find_spec(name="opentelemetry.instrumentation.fastapi"):
        from opentelemetry.instrumentation.fastapi import (  # pylint: disable=import-outside-toplevel # noqa: PLC0415
            FastAPIInstrumentor,
        )

        excluded_urls_str: str | None = None if len(config.excluded_urls) == 0 else ",".join(config.excluded_urls)
        FastAPIInstrumentor.instrument_app(  # pyright: ignore[reportUnknownMemberType]
            app=application.get_asgi_app(),
            tracer_provider=tracer_provider,
            meter_provider=meter_provider,
            excluded_urls=excluded_urls_str,
        )


def instrument_aiohttp(
    application: ApplicationAbstractProtocol,  # pylint: disable=unused-argument
    config: OpenTelemetryConfig,  # pylint: disable=unused-argument
    meter_provider: MeterProvider,
    tracer_provider: TracerProvider,
) -> None:
    """Instrument the Aiohttp application.

    Args:
        application (ApplicationAbstractProtocol): The application.
        config (OpenTelemetryConfig): The configuration.
        meter_provider (MeterProvider): The meter provider.
        tracer_provider (TracerProvider): The tracer provider.

    Returns:
        None
    """
    if find_spec(name="aiohttp") and find_spec(name="opentelemetry.instrumentation.aiohttp_client"):
        from opentelemetry.instrumentation.aiohttp_client import (  # pylint: disable=import-outside-toplevel # noqa: PLC0415
            AioHttpClientInstrumentor,
        )

        AioHttpClientInstrumentor().instrument(  # pyright: ignore[reportUnknownMemberType]
            tracer_provider=tracer_provider,
            meter_provider=meter_provider,
        )


def instrument_aio_pika(
    application: ApplicationAbstractProtocol,  # pylint: disable=unused-argument
    config: OpenTelemetryConfig,  # pylint: disable=unused-argument
    meter_provider: MeterProvider,
    tracer_provider: TracerProvider,
) -> None:
    """Instrument the AioPika application."""
    if find_spec(name="aio_pika") and find_spec(name="opentelemetry.instrumentation.aio_pika"):
        from opentelemetry.instrumentation.aio_pika import (  # pylint: disable=import-outside-toplevel # noqa: PLC0415
            AioPikaInstrumentor,
        )

        AioPikaInstrumentor().instrument(  # pyright: ignore[reportUnknownMemberType]
            tracer_provider=tracer_provider,
            meter_provider=meter_provider,
        )


INSTRUMENTS: list[Callable[..., Any]] = [instrument_fastapi, instrument_aiohttp, instrument_aio_pika]

__all__: list[str] = ["INSTRUMENTS"]



"""Provides the FastAPIBuilder class."""

from typing import Any, NamedTuple, Self

from fastapi import APIRouter, FastAPI
from fastapi.middleware import Middleware
from fastapi.middleware.cors import CORSMiddleware

from fastapi_factory_utilities.core.app.config import RootConfig


class MiddlewareArgs(NamedTuple):
    """Middleware arguments."""

    middleware_class: type[Middleware]
    kwargs: dict[str, Any]


class FastAPIBuilder:
    """FastAPI builder."""

    def __init__(self, root_config: RootConfig) -> None:
        """Instantiate the FastAPIBuilder."""
        self._root_config: RootConfig = root_config
        self._base_router: APIRouter = APIRouter()
        self._middleware_list: list[MiddlewareArgs] = []

    def add_api_router(self, router: APIRouter, without_resource_path: bool = False) -> Self:
        """Add the API router to the FastAPI application.

        Args:
            router: The API router.
            without_resource_path: If True, the resource path will not be added to the router.

        Returns:
            Self: The FastAPI builder.
        """
        self._base_router.include_router(
            router=router, prefix=self._root_config.application.root_path if not without_resource_path else ""
        )
        return self

    def add_middleware(self, middleware_class: type[Middleware], **kwargs: Any) -> Self:
        """Add a middleware to the FastAPI application.

        Args:
            middleware_class: The middleware class.
            *args: The middleware arguments.
            **kwargs: The middleware keyword arguments.

        Returns:
            Self: The FastAPI builder.
        """
        self._middleware_list.append(MiddlewareArgs(middleware_class=middleware_class, kwargs=kwargs))
        return self

    def build(self, lifespan: Any) -> FastAPI:
        """Build the FastAPI application.

        Returns:
            FastAPI: The FastAPI application.
        """
        fastapi = FastAPI(
            title=self._root_config.application.service_name,
            description="",
            version=self._root_config.application.version,
            lifespan=lifespan,
        )

        fastapi.add_middleware(
            middleware_class=CORSMiddleware,
            allow_origins=self._root_config.cors.allow_origins,
            allow_credentials=self._root_config.cors.allow_credentials,
            allow_methods=self._root_config.cors.allow_methods,
            allow_headers=self._root_config.cors.allow_headers,
        )

        for middleware_args in self._middleware_list:
            fastapi.add_middleware(
                middleware_class=middleware_args.middleware_class,  # type: ignore
                **middleware_args.kwargs,  # pyright: ignore
            )

        fastapi.include_router(router=self._base_router)

        return fastapi



"""Provides status services."""

import datetime

from fastapi import Request
from reactivex import Subject
from structlog.stdlib import BoundLogger, get_logger

from .enums import ComponentTypeEnum, HealthStatusEnum, ReadinessStatusEnum
from .exceptions import ComponentRegistrationError
from .health_calculator_strategies import (
    HealthCalculatorStrategy,
    HealthSimpleCalculatorStrategy,
)
from .readiness_calculator_strategies import (
    ReadinessCalculatorStrategy,
    ReadinessSimpleCalculatorStrategy,
)
from .types import (
    ComponentInstanceKey,
    ComponentInstanceType,
    Status,
    StatusUpdateEvent,
)

logger: BoundLogger = get_logger(__package__)


class StatusService:
    """Status service.

    It's responsible for managing the status of the components and determine the health and readiness status of the
    application.
    """

    def __init__(
        self,
        health_calculator_strategy: type[HealthCalculatorStrategy] = HealthSimpleCalculatorStrategy,
        readiness_calculator_strategy: type[ReadinessCalculatorStrategy] = ReadinessSimpleCalculatorStrategy,
    ) -> None:
        """Initialize the status service."""
        # Status
        self._health_status: HealthStatusEnum = HealthStatusEnum.HEALTHY
        self._health_calculator_strategy: type[HealthCalculatorStrategy] = health_calculator_strategy
        self._readiness_status: ReadinessStatusEnum = ReadinessStatusEnum.NOT_READY
        self._readiness_calculator_strategy: type[ReadinessCalculatorStrategy] = readiness_calculator_strategy
        # Components
        self._components: dict[str, ComponentInstanceType] = {}
        self._components_status: dict[ComponentInstanceKey, Status] = {}
        self._components_subjects: dict[ComponentInstanceKey, Subject[Status]] = {}
        # Observers
        self._status_subject: Subject[StatusUpdateEvent] = Subject()

    def _compute_status(self) -> None:
        """Compute the status."""
        # Health
        previous_health: HealthStatusEnum = self._health_status
        new_health: HealthStatusEnum = self._health_calculator_strategy(
            components_status=self._components_status
        ).calculate()
        # Readiness
        previous_readiness: ReadinessStatusEnum = self._readiness_status
        new_readiness: ReadinessStatusEnum = self._readiness_calculator_strategy(
            components_status=self._components_status
        ).calculate()
        # Update the status if needed
        if previous_health != new_health or previous_readiness != new_readiness:
            logger.info(
                "Status updated: health=%s, readiness=%s",
                new_health,
                new_readiness,
            )
            self._health_status = new_health
            self._readiness_status = new_readiness
            self._status_subject.on_next(
                StatusUpdateEvent(
                    health_status=new_health,
                    previous_health_status=previous_health,
                    readiness_status=new_readiness,
                    previous_readiness_status=previous_readiness,
                    triggered_at=datetime.datetime.now(),
                )
            )

    def get_status(self) -> Status:
        """Get the status.

        Returns:
            Status: The status.
        """
        return Status(
            health=self._health_status,
            readiness=self._readiness_status,
        )

    def get_components_status_by_type(self) -> dict[ComponentTypeEnum, dict[ComponentInstanceKey, Status]]:
        """Get the components status.

        Returns:
            dict[str, dict[str, Status]]: The components status.
        """
        result: dict[ComponentTypeEnum, dict[ComponentInstanceKey, Status]] = {}
        for component_instance in self._components.values():
            component_type: ComponentTypeEnum = component_instance.component_type
            if component_type not in result:
                result[component_type] = {}
            result[component_type][component_instance.key] = self._components_status[component_instance.key]
        return result

    def _on_next_for_component_instance(self, component_instance: ComponentInstanceType, event: Status) -> None:
        """On next subscribe for all component instances updates.

        Args:
            component_instance (ComponentInstanceType): The component instance.
            event (Status): The status event.

        Raises:
            ComponentRegistrationError: If the component instance is not registered.
        """
        # Check if the component instance is registered
        if component_instance.key not in self._components_status:
            raise ComponentRegistrationError(component_instance=component_instance)

        # Update the component instance status
        self._components_status[component_instance.key] = Status(
            health=event["health"],
            readiness=event["readiness"],
        )

        # Compute the status
        self._compute_status()

    def _register_component_instance_internaly(self, component_instance: ComponentInstanceType) -> None:
        """Register the component instance internally.

        Args:
            component_instance (ComponentInstanceType): The component instance.

        Raises:
            ComponentRegistrationError: If the component instance is already registered.
        """
        # Check if the component instance is already registered
        if component_instance.key in self._components:
            raise ComponentRegistrationError(component_instance=component_instance)
        # Register the component instance
        self._components[component_instance.key] = component_instance
        # Register the component instance status
        self._components_status[component_instance.key] = Status(
            health=HealthStatusEnum.HEALTHY,
            readiness=ReadinessStatusEnum.NOT_READY,
        )

    def _create_and_subscribe_to_component_instance_subject(
        self, component_instance: ComponentInstanceType
    ) -> Subject[Status]:
        """Create and subscribe to the component instance subject.

        Args:
            component_instance (ComponentInstanceType): The component instance.

        Returns:
            Subject[Status]: The observer.

        Raises:
            ComponentRegistrationError: If the component instance is already registered.
        """

        # Setup the subject and subscribe to the subject
        def on_next(event: Status) -> None:
            """Currying the on_next method for the component instance and delegate to the service method."""
            self._on_next_for_component_instance(
                component_instance=component_instance,
                event=event,
            )

        subject: Subject[Status] = Subject()
        # Check if the component instance is already registered
        if component_instance.key in self._components_subjects:
            raise ComponentRegistrationError(component_instance=component_instance)
        # Register the component instance subject
        self._components_subjects[component_instance.key] = subject
        subject.subscribe(on_next=on_next)

        return subject

    def register_component_instance(
        self,
        component_instance: ComponentInstanceType,
    ) -> Subject[Status]:
        """Register the component instance to the status service.

        It will create a subject for the component instance to share status update and
        the status service will subscribe automatically to it.

        Args:
            component_instance (ComponentInstanceType): The component instance.

        Returns:
            Subject[Status]: The observer.

        Raises:
            ComponentRegistrationError: If the component instance is already registered.
        """
        self._register_component_instance_internaly(component_instance=component_instance)
        subject: Subject[Status] = self._create_and_subscribe_to_component_instance_subject(
            component_instance=component_instance
        )
        logger.debug(
            "Component instance registered to the status service key=%s",
            component_instance.key,
            component_instance=component_instance,
        )
        return subject


def depends_status_service(request: Request) -> StatusService:
    """Get the status service, through fastapi depends."""
    return request.app.state.status_service



"""Provides enums for Kratos."""

from enum import StrEnum


class AuthenticatorAssuranceLevelEnum(StrEnum):
    """Enum for Authenticator Assurance Level (AAL)."""

    AAL1 = "aal1"
    AAL2 = "aal2"
    AAL3 = "aal3"



"""Provides the abstract class for the Aiopika plugin."""

from abc import ABC
from typing import Self

from aio_pika.abc import AbstractChannel, AbstractRobustConnection

from .exceptions import AiopikaPluginBaseError, AiopikaPluginConnectionNotProvidedError


class AbstractAiopikaResource(ABC):
    """Abstract class for the Aiopika resource."""

    def __init__(self) -> None:
        """Initialize the Aiopika resource."""
        self._robust_connection: AbstractRobustConnection | None = None
        self._channel: AbstractChannel | None = None

    def set_robust_connection(self, robust_connection: AbstractRobustConnection) -> Self:
        """Set the robust connection."""
        self._robust_connection = robust_connection
        return self

    def set_channel(self, channel: AbstractChannel) -> Self:
        """Set the channel."""
        self._channel = channel
        return self

    async def _acquire_channel(self) -> AbstractChannel:
        """Acquire the channel."""
        if self._robust_connection is None:
            raise AiopikaPluginConnectionNotProvidedError(
                message="Robust connection not provided.",
            )
        if self._channel is None:
            try:
                self._channel = await self._robust_connection.channel()
            except Exception as exception:
                raise AiopikaPluginBaseError(
                    message="Failed to acquire the channel.",
                ) from exception
        return self._channel

    async def setup(self) -> Self:
        """Setup the Aiopika resource."""
        if self._channel is None:
            await self._acquire_channel()
        return self



"""Provides the configuration model for the OpenTelemetry plugin."""

from enum import StrEnum
from typing import Annotated

from pydantic import BaseModel, ConfigDict, Field, UrlConstraints
from pydantic_core import Url


class ProtocolEnum(StrEnum):
    """Defines the protocol enum for OpenTelemetry."""

    OTLP_GRPC = "otlp_grpc"
    OTLP_HTTP = "otlp_http"


class OpenTelemetryMeterConfig(BaseModel):
    """Provides the configuration model for the OpenTelemetry meter as sub-model."""

    # Constants for time in milliseconds and seconds
    ONE_MINUTE_IN_MILLIS: int = 60000
    ONE_SECOND_IN_MILLIS: int = 1000

    model_config = ConfigDict(frozen=True, extra="forbid")

    reader_interval_millis: float = Field(
        default=ONE_SECOND_IN_MILLIS,
        description="The interval in miliseconds to read and export metrics.",
    )

    reader_timeout_millis: float = Field(
        default=ONE_SECOND_IN_MILLIS,
        description="The timeout in miliseconds for the reader.",
    )


class OpenTelemetryTracerConfig(BaseModel):
    """Provides the configuration model for the OpenTelemetry tracer as sub-model."""

    # Constants for time in milliseconds and seconds
    FIVE_SECONDS_IN_MILLIS: int = 5000
    THIRTY_SECONDS_IN_MILLIS: int = 30000

    # Default values for the OpenTelemetry tracer
    DEFAULT_OTEL_TRACER_MAX_QUEUE_SIZE: int = 2048
    DEFAULT_OTEL_TRACER_MAX_EXPORT_BATCH_SIZE: int = 512

    model_config = ConfigDict(frozen=True, extra="forbid")

    max_queue_size: int = Field(
        default=DEFAULT_OTEL_TRACER_MAX_QUEUE_SIZE,
        description="The maximum queue size for the tracer.",
    )
    max_export_batch_size: int = Field(
        default=DEFAULT_OTEL_TRACER_MAX_EXPORT_BATCH_SIZE,
        description="The maximum export batch size for the tracer.",
    )
    schedule_delay_millis: int = Field(
        default=FIVE_SECONDS_IN_MILLIS,
        description="The schedule delay in miliseconds for the tracer.",
    )
    export_timeout_millis: int = Field(
        default=THIRTY_SECONDS_IN_MILLIS,
        description="The export timeout in miliseconds for the tracer.",
    )


class OpenTelemetryConfig(BaseModel):
    """Provides the configuration model for the OpenTelemetry plugin."""

    # Constants for time in milliseconds and seconds
    TEN_SECONDS_IN_SECONDS: int = 10

    # Default value for the collector endpoint
    DEFAULT_COLLECTOR_ENDPOINT: str = "http://localhost:4318"

    model_config = ConfigDict(frozen=True, extra="forbid")

    activate: bool = Field(
        default=False,
        description="Whether to activate the OpenTelemetry collector export.",
    )
    endpoint: Annotated[Url, UrlConstraints(allowed_schemes=["http", "https"])] = Field(
        default=Url(url=DEFAULT_COLLECTOR_ENDPOINT),
        description="The collector endpoint.",
    )

    protocol: ProtocolEnum | None = Field(
        default=None,
        description="The protocol to use for the collector.",
    )

    timeout: int = Field(
        default=TEN_SECONDS_IN_SECONDS,
        description="The timeout in seconds for the collector.",
    )

    closing_timeout: int = Field(
        default=TEN_SECONDS_IN_SECONDS,
        description="The closing timeout in seconds for the collector.",
    )

    meter_config: OpenTelemetryMeterConfig | None = Field(
        default_factory=OpenTelemetryMeterConfig,
        description="The meter configuration.",
    )

    tracer_config: OpenTelemetryTracerConfig | None = Field(
        default_factory=OpenTelemetryTracerConfig,
        description="The tracer configuration.",
    )

    excluded_urls: list[str] = Field(
        default_factory=list,
        description="The excluded URLs for both the metrics and traces.",
    )



"""Provides the configurations for the Taskiq plugin."""

from typing import ClassVar

from pydantic import BaseModel, ConfigDict


class RedisCredentialsConfig(BaseModel):
    """Redis credentials config."""

    model_config: ClassVar[ConfigDict] = ConfigDict(frozen=True, extra="forbid")
    url: str



"""Provides a function to setup the logging configuration."""

import logging
import sys
from enum import StrEnum, auto
from typing import Annotated, Any

import structlog
from pydantic import BaseModel, BeforeValidator
from structlog.types import EventDict

_logger = structlog.getLogger(__package__)


def ensure_logging_level(level: Any) -> int:
    """Ensure the logging level.

    Args:
        level (Any): The logging level.

    Returns:
        int: The logging level.
    """
    if isinstance(level, int):
        return level

    if isinstance(level, str):
        try:
            return getattr(logging, str(level).upper())
        except AttributeError as exception:
            raise ValueError(f"Invalid logging level: {level}") from exception

    raise ValueError(f"Invalid logging level: {level}")


class LoggingConfig(BaseModel):
    """Logging configuration."""

    name: str
    level: Annotated[int, BeforeValidator(ensure_logging_level)]


class LogModeEnum(StrEnum):
    """Defines the possible logging modes."""

    CONSOLE = auto()
    JSON = auto()


# https://github.com/hynek/structlog/issues/35#issuecomment-591321744
def _rename_event_key(_: Any, __: Any, event_dict: EventDict) -> EventDict:  # pylint: disable=invalid-name
    """Renames the `event` key to `message` in the event dictionary.

    Log entries keep the text message in the `event` field, but Datadog
    uses the `message` field. This processor moves the value from one field to
    the other.
    See https://github.com/hynek/structlog/issues/35#issuecomment-591321744
    """
    event_dict["message"] = event_dict.pop("event")
    return event_dict


def clean_uvicorn_logger() -> None:
    """Cleans the uvicorn loggers."""
    for logger_name in ["uvicorn", "uvicorn.error", "uvicorn.access"]:
        # Clear the log handlers for uvicorn loggers, and enable propagation
        # so the messages are caught by our root logger and formatted correctly
        # by structlog
        logging.getLogger(logger_name).handlers.clear()
        logging.getLogger(logger_name).propagate = True


def _drop_color_message_key(_: Any, __: Any, event_dict: EventDict) -> EventDict:  # pylint: disable=invalid-name
    """Cleans the `color_message` key from the event dictionary.

    Uvicorn logs the message a second time in the extra `color_message`, but we don't
    need it. This processor drops the key from the event dict if it exists.
    """
    event_dict.pop("color_message", None)
    return event_dict


def setup_log(
    mode: LogModeEnum = LogModeEnum.CONSOLE, log_level: str = "DEBUG", logging_config: list[LoggingConfig] | None = None
) -> None:
    """Prepares the logging configuration.

    Args:
        mode (LogMode): The logging mode to use.
        log_level (str): The log level to use.
        logging_config (List[LoggingConfig], optional): The logging configuration. Defaults to None.

    Returns:
        None
    """
    processors: list[structlog.typing.Processor] = [
        structlog.stdlib.add_logger_name,
        structlog.stdlib.add_log_level,
        structlog.processors.TimeStamper(fmt="iso"),
        structlog.processors.UnicodeDecoder(),
        structlog.stdlib.PositionalArgumentsFormatter(),
        structlog.stdlib.ExtraAdder(),
        structlog.processors.CallsiteParameterAdder(
            parameters={
                structlog.processors.CallsiteParameter.MODULE: True,
                structlog.processors.CallsiteParameter.FUNC_NAME: True,
                structlog.processors.CallsiteParameter.LINENO: True,
            }
        ),
        _drop_color_message_key,
    ]

    log_renderer: structlog.dev.ConsoleRenderer | structlog.processors.JSONRenderer
    match mode:
        case LogModeEnum.CONSOLE:
            log_renderer = structlog.dev.ConsoleRenderer(
                exception_formatter=structlog.dev.RichTracebackFormatter(),
            )
        case LogModeEnum.JSON:
            # We rename the `event` key to `message` only in JSON logs,
            # as Datadog looks for the
            # `message` key but the pretty ConsoleRenderer looks for `event`
            processors.append(_rename_event_key)
            # Format the exception only for JSON logs, as we want
            # to pretty-print them when
            # using the ConsoleRenderer
            processors.append(
                structlog.processors.dict_tracebacks,
            )
            log_renderer = structlog.processors.JSONRenderer()

    # Remove all existing loggers
    structlog.reset_defaults()
    structlog_processors: list[structlog.typing.Processor] = processors.copy()
    structlog_processors.append(structlog.stdlib.ProcessorFormatter.wrap_for_formatter)
    structlog.configure(
        processors=structlog_processors,
        logger_factory=structlog.stdlib.LoggerFactory(),
        cache_logger_on_first_use=True,
    )

    formatter = structlog.stdlib.ProcessorFormatter(
        # These run ONLY on `logging` entries that do NOT originate within
        # structlog.
        foreign_pre_chain=processors,
        # These run on ALL entries after the pre_chain is done.
        processors=[
            # Remove _record & _from_structlog.
            structlog.stdlib.ProcessorFormatter.remove_processors_meta,
            log_renderer,
        ],
    )

    handler = logging.StreamHandler()
    # Use OUR `ProcessorFormatter` to format all `logging` entries.
    handler.setFormatter(formatter)
    root_logger: logging.Logger = logging.getLogger()
    root_logger.addHandler(handler)
    root_logger.setLevel(log_level.upper())

    for logging_config_item in logging_config or []:
        logger = logging.getLogger(logging_config_item.name)
        logger.setLevel(logging_config_item.level)

    def handle_exception(exc_type: Any, exc_value: Any, exc_traceback: Any) -> None:
        """Handle uncaught exceptions.

        Log any uncaught exception instead of letting it be printed by Python
        (but leave KeyboardInterrupt untouched to allow users to Ctrl+C to stop)
        See https://stackoverflow.com/a/16993115/3641865
        """
        if issubclass(exc_type, KeyboardInterrupt):
            sys.__excepthook__(exc_type, exc_value, exc_traceback)
            return

        _logger.error("Uncaught exception", exc_info=(exc_type, exc_value, exc_traceback))

    sys.excepthook = handle_exception



"""Provides helper functions for the OpenTelemetry plugin."""

from collections.abc import Callable
from functools import wraps
from typing import ParamSpec, TypeVar

from opentelemetry import trace
from opentelemetry.context import Context
from opentelemetry.trace import SpanKind
from opentelemetry.util import types

Param = ParamSpec("Param")
RetType = TypeVar("RetType")  # pylint: disable=invalid-name


def trace_span(
    name: str | None = None,
    context: Context | None = None,
    kind: SpanKind = SpanKind.INTERNAL,
    attributes: types.Attributes = None,
) -> Callable[[Callable[Param, RetType]], Callable[Param, RetType]]:
    """Decorator to trace a function using OpenTelemetry."""

    def decorator(func: Callable[Param, RetType]) -> Callable[Param, RetType]:
        @wraps(wrapped=func)
        def wrapper(*args: Param.args, **kwargs: Param.kwargs) -> RetType:
            # Get Tracer from the instrumented function's module
            tracer: trace.Tracer = trace.get_tracer(instrumenting_module_name=func.__module__)
            # Use the function's name as the span name if no name is provided
            trace_name: str = name if name is not None else func.__name__
            # Start a span with the provided name
            with tracer.start_as_current_span(
                name=trace_name,
                kind=kind,
                context=context,
                attributes=attributes,
            ):
                return func(*args, **kwargs)

        return wrapper

    return decorator



"""Oriented Data Model (ODM) plugin package."""

from logging import INFO, Logger, getLogger
from typing import Any, Self

from beanie import Document, init_beanie  # pyright: ignore[reportUnknownVariableType]
from motor.motor_asyncio import AsyncIOMotorClient, AsyncIOMotorDatabase
from reactivex import Subject
from structlog.stdlib import BoundLogger, get_logger

from fastapi_factory_utilities.core.plugins.abstracts import PluginAbstract
from fastapi_factory_utilities.core.protocols import ApplicationAbstractProtocol
from fastapi_factory_utilities.core.services.status.enums import (
    ComponentTypeEnum,
    HealthStatusEnum,
    ReadinessStatusEnum,
)
from fastapi_factory_utilities.core.services.status.services import StatusService
from fastapi_factory_utilities.core.services.status.types import (
    ComponentInstanceType,
    Status,
)

from .builder import ODMBuilder
from .configs import ODMConfig
from .depends import depends_odm_client, depends_odm_database
from .documents import BaseDocument
from .exceptions import OperationError, UnableToCreateEntityDueToDuplicateKeyError
from .helpers import PersistedEntity
from .repositories import AbstractRepository

_logger: BoundLogger = get_logger()


class ODMPlugin(PluginAbstract):
    """ODM plugin."""

    def __init__(
        self, document_models: list[type[Document]] | None = None, odm_config: ODMConfig | None = None
    ) -> None:
        """Initialize the ODM plugin."""
        super().__init__()
        self._component_instance: ComponentInstanceType | None = None
        self._monitoring_subject: Subject[Status] | None = None
        self._document_models: list[type[Document]] | None = document_models
        self._odm_config: ODMConfig | None = odm_config
        self._odm_client: AsyncIOMotorClient[Any] | None = None
        self._odm_database: AsyncIOMotorDatabase[Any] | None = None

    def set_application(self, application: ApplicationAbstractProtocol) -> Self:
        """Set the application."""
        self._document_models = self._document_models or application.ODM_DOCUMENT_MODELS
        return super().set_application(application)

    def on_load(self) -> None:
        """Actions to perform on load for the ODM plugin."""
        # Configure the pymongo logger to INFO level

        pymongo_logger: Logger = getLogger("pymongo")
        pymongo_logger.setLevel(INFO)
        _logger.debug("ODM plugin loaded.")

    def _setup_status(self) -> None:
        assert self._application is not None
        status_service: StatusService = self._application.get_status_service()
        self._component_instance = ComponentInstanceType(
            component_type=ComponentTypeEnum.DATABASE, identifier="MongoDB"
        )
        self._monitoring_subject = status_service.register_component_instance(
            component_instance=self._component_instance
        )

    async def _setup_beanie(self) -> None:
        assert self._application is not None
        assert self._odm_database is not None
        assert self._document_models is not None
        assert self._monitoring_subject is not None
        # TODO: Find a better way to initialize beanie with the document models of the concrete application
        # through an hook in the application, a dynamis import ?
        try:
            await init_beanie(
                database=self._odm_database,
                document_models=self._document_models,
            )
        except Exception as exception:  # pylint: disable=broad-except
            _logger.error(f"ODM plugin failed to start. {exception}")
            # TODO: Report the error to the status_service
            # this will report the application as unhealthy
            self._monitoring_subject.on_next(
                value=Status(health=HealthStatusEnum.UNHEALTHY, readiness=ReadinessStatusEnum.NOT_READY)
            )

    async def on_startup(self) -> None:
        """Actions to perform on startup for the ODM plugin."""
        assert self._application is not None
        self._setup_status()
        assert self._monitoring_subject is not None
        assert self._component_instance is not None

        try:
            odm_factory: ODMBuilder = ODMBuilder(application=self._application, odm_config=self._odm_config).build_all()
            await odm_factory.wait_ping()
            self._odm_database = odm_factory.odm_database
            self._odm_client = odm_factory.odm_client
        except Exception as exception:  # pylint: disable=broad-except
            _logger.error(f"ODM plugin failed to start. {exception}")
            # TODO: Report the error to the status_service
            # this will report the application as unhealthy
            self._monitoring_subject.on_next(
                value=Status(health=HealthStatusEnum.UNHEALTHY, readiness=ReadinessStatusEnum.NOT_READY)
            )
            return None

        if self._odm_database is None or self._odm_client is None:
            _logger.error(
                f"ODM plugin failed to start. Database: {odm_factory.odm_database} - Client: {odm_factory.odm_client}"
            )
            # TODO: Report the error to the status_service
            # this will report the application as unhealthy
            self._monitoring_subject.on_next(
                value=Status(health=HealthStatusEnum.UNHEALTHY, readiness=ReadinessStatusEnum.NOT_READY)
            )
            return None

        self._add_to_state(key="odm_client", value=odm_factory.odm_client)
        self._add_to_state(key="odm_database", value=odm_factory.odm_database)

        await self._setup_beanie()

        _logger.info(
            f"ODM plugin started. Database: {self._odm_database.name} - "
            f"Client: {self._odm_client.address} - "
            f"Document models: {self._application.ODM_DOCUMENT_MODELS}"
        )

        self._monitoring_subject.on_next(
            value=Status(health=HealthStatusEnum.HEALTHY, readiness=ReadinessStatusEnum.READY)
        )

    async def on_shutdown(self) -> None:
        """Actions to perform on shutdown for the ODM plugin."""
        if self._odm_client is not None:
            self._odm_client.close()
        _logger.debug("ODM plugin shutdown.")


__all__: list[str] = [
    "AbstractRepository",
    "BaseDocument",
    "OperationError",
    "PersistedEntity",
    "UnableToCreateEntityDueToDuplicateKeyError",
    "depends_odm_client",
    "depends_odm_database",
]



"""Python exceptions for the Kratos service."""

from fastapi_factory_utilities.core.exceptions import FastAPIFactoryUtilitiesError


class KratosError(FastAPIFactoryUtilitiesError):
    """Base class for all exceptions raised by the Kratos service."""


class KratosOperationError(KratosError):
    """Exception raised when a Kratos operation fails."""


class KratosSessionInvalidError(KratosOperationError):
    """Exception raised when a Kratos session is invalid."""



"""Provides services for books."""

from typing import ClassVar
from uuid import UUID

from fastapi import Request
from opentelemetry import metrics

from fastapi_factory_utilities.core.plugins.odm_plugin.depends import (
    depends_odm_database,
)
from fastapi_factory_utilities.core.plugins.opentelemetry_plugin.helpers import (
    trace_span,
)
from fastapi_factory_utilities.example.entities.books import (
    BookEntity,
    BookName,
    BookType,
)
from fastapi_factory_utilities.example.models.books.repository import BookRepository


class BookService:
    """Provides services for books."""

    book_store: ClassVar[dict[UUID, BookEntity]] = {}

    # Metrics Definitions
    METER_COUNTER_BOOK_GET_NAME: str = "book_get"
    METER_COUNTER_BOOK_ADD_NAME: str = "book_add"
    METER_COUNTER_BOOK_REMOVE_NAME: str = "book_remove"
    METER_COUNTER_BOOK_UPDATE_NAME: str = "book_update"
    # ====================

    meter: metrics.Meter = metrics.get_meter(__name__)

    METER_COUNTER_BOOK_GET: metrics.Counter = meter.create_counter(
        name=METER_COUNTER_BOOK_GET_NAME, description="The number of books retrieved."
    )
    METER_COUNTER_BOOK_ADD: metrics.Counter = meter.create_counter(
        name=METER_COUNTER_BOOK_ADD_NAME,
        description="The number of books added.",
    )
    METER_COUNTER_BOOK_REMOVE: metrics.Counter = meter.create_counter(
        name=METER_COUNTER_BOOK_REMOVE_NAME,
        description="The number of books removed.",
    )
    METER_COUNTER_BOOK_UPDATE: metrics.Counter = meter.create_counter(
        name=METER_COUNTER_BOOK_UPDATE_NAME,
        description="The number of books updated.",
    )

    def __init__(
        self,
        book_repository: BookRepository,
    ) -> None:
        """Initialize the service.

        Args:
            book_repository (injector.Inject[BookRepository]): The book repository.
            meter (metrics.Meter, optional): The meter. Defaults to None use for testing purpose.

        Raises:
            ValueError: If the book already exists.

        """
        self.book_repository: BookRepository = book_repository

        # Build the book store if it is empty
        if len(self.book_store) == 0:
            self.build_book_store()

    @classmethod
    def build_default_book_store(cls) -> list[BookEntity]:
        """Build the default book store."""
        return [
            BookEntity(title=BookName("Book 1"), book_type=BookType.FANTASY),
            BookEntity(title=BookName("Book 2"), book_type=BookType.MYSTERY),
            BookEntity(title=BookName("Book 3"), book_type=BookType.SCIENCE_FICTION),
        ]

    @classmethod
    def build_book_store(cls, books: list[BookEntity] | None = None) -> None:
        """Build the book store.

        Args:
            books (list[BookEntity], optional): The books to add. Defaults to None.
        """
        if books is None:
            books = cls.build_default_book_store()

        cls.book_store = {book.id: book for book in books}

    @trace_span(name="Add Book")
    def add_book(self, book: BookEntity) -> None:
        """Add a book.

        Args:
            book (BookEntity): The book to add

        Raises:
            ValueError: If the book already exists.
        """
        if book.id in self.book_store:
            raise ValueError(f"Book with id {book.id} already exists.")

        self.book_store[book.id] = book

        self.METER_COUNTER_BOOK_ADD.add(amount=1)

    def get_book(self, book_id: UUID) -> BookEntity:
        """Get a book.

        Args:
            book_id (UUID): The book id

        Returns:
            BookEntity: The book

        Raises:
            ValueError: If the book does not exist.
        """
        if book_id not in self.book_store:
            raise ValueError(f"Book with id {book_id} does not exist.")

        self.METER_COUNTER_BOOK_GET.add(amount=1, attributes={"book_count": 1})

        return self.book_store[book_id]

    def get_all_books(self) -> list[BookEntity]:
        """Get all books.

        Returns:
            list[BookEntity]: All books
        """
        self.METER_COUNTER_BOOK_GET.add(amount=1, attributes={"book_count": len(self.book_store)})
        return list(self.book_store.values())

    @trace_span(name="Remove Book")
    def remove_book(self, book_id: UUID) -> None:
        """Remove a book.

        Args:
            book_id (UUID): The book id

        Raises:
            ValueError: If the book does not exist.
        """
        if book_id not in self.book_store:
            raise ValueError(f"Book with id {book_id} does not exist.")

        del self.book_store[book_id]

        self.METER_COUNTER_BOOK_REMOVE.add(amount=1)

    @trace_span(name="Update Book")
    def update_book(self, book: BookEntity) -> None:
        """Update a book.

        Args:
            book (BookEntity): The book to update

        Raises:
            ValueError: If the book does not exist.
        """
        if book.id not in self.book_store:
            raise ValueError(f"Book with id {book.id} does not exist.")

        self.book_store[book.id] = book

        self.METER_COUNTER_BOOK_UPDATE.add(amount=1)


def depends_book_service(request: Request) -> BookService:
    """Provide Book Service."""
    return BookService(book_repository=BookRepository(database=depends_odm_database(request=request)))



"""Provides the exceptions for the OpenTelemetryPlugin."""


class OpenTelemetryPluginBaseException(BaseException):
    """Base exception for the OpenTelemetryPlugin."""

    pass


class OpenTelemetryPluginConfigError(OpenTelemetryPluginBaseException):
    """Exception for the OpenTelemetryPlugin configuration."""

    pass



"""Abstracts for the plugins."""

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Self

if TYPE_CHECKING:
    from fastapi_factory_utilities.core.protocols import ApplicationAbstractProtocol


class PluginAbstract(ABC):
    """Abstract class for the plugins."""

    def __init__(self) -> None:
        """Initialize the plugin."""
        self._application: ApplicationAbstractProtocol | None = None

    def set_application(self, application: "ApplicationAbstractProtocol") -> Self:
        """Set the application."""
        self._application = application
        return self

    def _add_to_state(self, key: str, value: Any) -> None:
        """Add to the state."""
        assert self._application is not None
        setattr(self._application.get_asgi_app().state, key, value)

    @abstractmethod
    def on_load(self) -> None:
        """On load."""
        raise NotImplementedError

    @abstractmethod
    async def on_startup(self) -> None:
        """On startup."""
        raise NotImplementedError

    @abstractmethod
    async def on_shutdown(self) -> None:
        """On shutdown."""
        raise NotImplementedError



"""Provides the queue for the Aiopika plugin."""

from typing import ClassVar, Self

from aio_pika.abc import AbstractQueue, TimeoutType

from .abstract import AbstractAiopikaResource
from .exceptions import AiopikaPluginBaseError, AiopikaPluginQueueNotDeclaredError
from .exchange import Exchange


class Queue(AbstractAiopikaResource):
    """Queue."""

    DEFAULT_OPERATION_TIMEOUT: ClassVar[TimeoutType] = 10.0

    def __init__(  # pylint: disable=too-many-arguments # noqa: PLR0913
        self,
        name: str,
        exchange: Exchange,
        routing_key: str,
        durable: bool = True,
        auto_delete: bool = False,
        exclusive: bool = True,
        timeout: TimeoutType = DEFAULT_OPERATION_TIMEOUT,
    ) -> None:
        """Initialize the queue."""
        super().__init__()
        # Initialize the queue properties
        self._name: str = name
        self._routing_key: str = routing_key
        self._durable: bool = durable
        self._auto_delete: bool = auto_delete
        self._exclusive: bool = exclusive
        self._timeout: TimeoutType = timeout
        # Behavior properties
        self._exchange: Exchange = exchange
        self._queue: AbstractQueue | None = None

    @property
    def queue(self) -> AbstractQueue:
        """Get the Aiopika queue."""
        if self._queue is None:
            raise AiopikaPluginQueueNotDeclaredError(
                message="Queue not declared.",
                queue=self._name,
            )
        return self._queue

    async def _declare(self) -> Self:
        """Declare the queue."""
        try:
            self._queue = await self._channel.declare_queue(  # pyright: ignore
                name=self._name,
                durable=self._durable,
                auto_delete=self._auto_delete,
                exclusive=self._exclusive,
                timeout=self._timeout,
            )
        except Exception as exception:
            raise AiopikaPluginBaseError(
                message="Failed to declare the queue.",
            ) from exception
        return self

    async def _bind(self) -> Self:
        """Bind the queue to the exchange."""
        try:
            await self._queue.bind(  # pyright: ignore
                exchange=self._exchange.exchange,
                routing_key=self._routing_key,
                timeout=self._timeout,
            )
        except Exception as exception:
            raise AiopikaPluginBaseError(
                message="Failed to bind the queue to the exchange.",
            ) from exception
        return self

    async def setup(self) -> Self:
        """Setup the queue."""
        await super().setup()
        if self._queue is None:
            await self._declare()
        await self._bind()
        return self



"""Helper functions for ODM plugins."""

import datetime
import uuid

from pydantic import BaseModel, Field


class PersistedEntity(BaseModel):
    """Base class for persisted entities."""

    id: uuid.UUID = Field(default_factory=uuid.uuid4)

    revision_id: uuid.UUID | None = Field(default=None)
    created_at: datetime.datetime = Field(default_factory=datetime.datetime.now)
    updated_at: datetime.datetime = Field(default_factory=datetime.datetime.now)



"""Model for Book."""

from typing import Annotated

from beanie import Indexed  # pyright: ignore[reportUnknownVariableType]

from fastapi_factory_utilities.core.plugins.odm_plugin.documents import BaseDocument
from fastapi_factory_utilities.example.entities.books import BookName, BookType


class BookDocument(BaseDocument):
    """BookModel."""

    title: Annotated[BookName, Indexed(unique=True)]
    book_type: Annotated[BookType, Indexed()]

    class Settings(BaseDocument.Settings):
        """Meta class for BookModel."""

        collection: str = "books"



"""Kratos service module."""

from .exceptions import KratosOperationError, KratosSessionInvalidError
from .objects import KratosSessionObject
from .services import KratosService, depends_kratos_service

__all__: list[str] = [
    "KratosOperationError",
    "KratosService",
    "KratosSessionInvalidError",
    "KratosSessionObject",
    "depends_kratos_service",
]



"""Provide the ApplicationGenericBuilder class."""

from typing import Any, Generic, Self, TypeVar, get_args

from fastapi_factory_utilities.core.app.config import GenericConfigBuilder, RootConfig
from fastapi_factory_utilities.core.app.fastapi_builder import FastAPIBuilder
from fastapi_factory_utilities.core.plugins import PluginAbstract
from fastapi_factory_utilities.core.utils.log import LogModeEnum, setup_log
from fastapi_factory_utilities.core.utils.uvicorn import UvicornUtils

from .application import ApplicationAbstract

T = TypeVar("T", bound=ApplicationAbstract)


class ApplicationGenericBuilder(Generic[T]):
    """Application generic builder."""

    def __init__(self, plugins: list[PluginAbstract] | None = None) -> None:
        """Instanciate the ApplicationGenericBuilder."""
        self._uvicorn_utils: UvicornUtils | None = None
        self._root_config: RootConfig | None = None
        self._plugins: list[PluginAbstract] = plugins or []
        self._fastapi_builder: FastAPIBuilder | None = None
        generic_args: tuple[Any, ...] = get_args(self.__orig_bases__[0])  # type: ignore
        self._application_class: type[T] = generic_args[0]

    def add_plugin_to_activate(self, plugin: PluginAbstract) -> Self:
        """Add a plugin to activate.

        Args:
            plugin (PluginsEnum): The plugin to activate.

        Returns:
            Self: The builder.
        """
        self._plugins.append(plugin)
        return self

    def add_config(self, config: RootConfig) -> Self:
        """Add the configuration to the builder.

        Args:
            config (RootConfig): The configuration.

        Returns:
            Self: The builder.
        """
        self._root_config = config
        return self

    def add_fastapi_builder(self, fastapi_builder: FastAPIBuilder) -> Self:
        """Add the FastAPI builder to the builder.

        Args:
            fastapi_builder (FastAPIBuilder): The FastAPI builder.

        Returns:
            Self: The builder.
        """
        self._fastapi_builder = fastapi_builder
        return self

    def _build_from_package_root_config(self) -> RootConfig:
        """Build the configuration from the package."""
        return GenericConfigBuilder[self._application_class.CONFIG_CLASS](  # type: ignore
            package_name=self._application_class.PACKAGE_NAME,
            config_class=self._application_class.CONFIG_CLASS,
        ).build()

    def build(self, **kwargs: Any) -> T:
        """Build the application.

        Args:
            **kwargs: The keyword arguments to pass to the application.

        Returns:
            T: The application.
        """
        # RootConfig
        self._root_config = self._root_config or self._build_from_package_root_config()
        # FastAPIBuilder
        self._fastapi_builder = self._fastapi_builder or FastAPIBuilder(root_config=self._root_config)

        application: T = self._application_class(
            root_config=self._root_config,
            plugins=self._plugins,
            fastapi_builder=self._fastapi_builder,
            **kwargs,
        )
        application.setup()
        return application

    def build_as_uvicorn_utils(self) -> UvicornUtils:
        """Build the application and provide UvicornUtils."""
        self._uvicorn_utils = UvicornUtils(app=self.build())
        return self._uvicorn_utils

    def build_and_serve(self) -> None:
        """Build the application and serve it with Uvicorn."""
        uvicorn_utils: UvicornUtils = self._uvicorn_utils or self.build_as_uvicorn_utils()

        assert self._root_config is not None
        setup_log(mode=LogModeEnum.CONSOLE, logging_config=self._root_config.logging)

        try:
            uvicorn_utils.serve()
        except KeyboardInterrupt:
            pass



"""Python exceptions for the Hydra service."""

from fastapi_factory_utilities.core.exceptions import FastAPIFactoryUtilitiesError


class HydraError(FastAPIFactoryUtilitiesError):
    """Base class for all exceptions raised by the Hydra service."""


class HydraOperationError(HydraError):
    """Exception raised when a Hydra operation fails."""


class HydraTokenInvalidError(HydraError):
    """Exception raised when a Hydra token is invalid."""



"""Provides the exceptions for the application factory."""

from ..exceptions import FastAPIFactoryUtilitiesError


class BaseApplicationException(BaseException):
    """Base application exception."""

    pass


class ConfigBuilderError(FastAPIFactoryUtilitiesError):
    """Application configuration factory exception."""

    def __init__(self, message: str, config_class: type, package: str, filename: str) -> None:
        """Instantiate the exception."""
        super().__init__(
            message=f"Unable to build the configuration for the package {package} and "
            + f"the file {filename} with the class {config_class} - {message}"
        )



"""Scheduler module for fastapi_factory_utilities.

This module provides components and utilities for scheduling tasks using Taskiq, FastAPI, and Redis.
It enables registration, configuration, and management of scheduled tasks in FastAPI applications.
"""

import asyncio
from collections.abc import Coroutine
from typing import Any, Self, cast

import taskiq_fastapi
from fastapi import FastAPI
from structlog.stdlib import get_logger
from taskiq import (
    AsyncBroker,
    AsyncTaskiqDecoratedTask,
    ScheduleSource,
    TaskiqScheduler,
)
from taskiq.api import run_receiver_task, run_scheduler_task
from taskiq.scheduler.created_schedule import CreatedSchedule
from taskiq.scheduler.scheduled_task import ScheduledTask
from taskiq_redis import (
    ListRedisScheduleSource,
    RedisAsyncResultBackend,
    RedisStreamBroker,
)

_logger = get_logger(__package__)


class SchedulerComponent:
    """Scheduler component."""

    NAME_SUFFIX: str = "tiktok_integration"

    def __init__(self) -> None:
        """Initialize the scheduler component."""
        self._result_backend: RedisAsyncResultBackend[Any] | None = None
        self._stream_broker: RedisStreamBroker | None = None
        self._scheduler: TaskiqScheduler | None = None
        self._scheduler_source: ListRedisScheduleSource | None = None
        self._dyn_task: AsyncTaskiqDecoratedTask[Any, Any] | None = None
        self._schedule_cron: ScheduledTask | None = None
        self._schedulers_tasks: dict[str, AsyncTaskiqDecoratedTask[Any, Any]] = {}

    def register_task(self, task: Coroutine[Any, Any, Any], task_name: str) -> None:
        """Register a task.

        Args:
            task: The task to register.
            task_name: The name of the task.

        Raises:
            ValueError: If the task is already registered.
            ValueError: If the stream broker is not initialized.
        """
        if self._stream_broker is None:
            raise ValueError("Stream broker is not initialized")

        if task_name in self._schedulers_tasks:
            raise ValueError(f"Task {task_name} already registered")

        self._schedulers_tasks[task_name] = self._stream_broker.register_task(task, task_name)  # type: ignore

    def get_task(self, task_name: str) -> AsyncTaskiqDecoratedTask[Any, Any]:
        """Get a task.

        Args:
            task_name: The name of the task.

        Returns:
            AsyncTaskiqDecoratedTask: The task.

        Raises:
            ValueError: If the task is not registered.
        """
        if task_name not in self._schedulers_tasks:
            raise ValueError(f"Task {task_name} not registered")
        return self._schedulers_tasks[task_name]

    def configure(self, redis_connection_string: str, app: FastAPI) -> Self:
        """Configure the scheduler component."""
        self._result_backend = RedisAsyncResultBackend(
            redis_url=redis_connection_string,
            prefix_str=f"velmios_taskiq_result_backend_{self.NAME_SUFFIX}",
            result_ex_time=120,
        )
        self._stream_broker = RedisStreamBroker(
            url=redis_connection_string,
            queue_name=f"velmios_taskiq_stream_broker_{self.NAME_SUFFIX}",
            consumer_group_name=f"velmios_taskiq_consumer_group_{self.NAME_SUFFIX}",
        ).with_result_backend(self._result_backend)

        taskiq_fastapi.populate_dependency_context(self._stream_broker, app)

        self._scheduler_source = ListRedisScheduleSource(
            url=redis_connection_string,
            prefix=f"velmios_taskiq_schedule_source_{self.NAME_SUFFIX}",
        )

        self._scheduler = TaskiqScheduler(
            broker=self._stream_broker,
            sources=[self._scheduler_source],
        )

        return self

    async def startup(self, app: FastAPI) -> None:
        """Start the scheduler."""
        if self._result_backend is None:
            raise ValueError("Result backend is not initialized")
        if self._stream_broker is None:
            raise ValueError("Stream broker is not initialized")
        if self._scheduler is None:
            raise ValueError("Scheduler is not initialized")
        if self._scheduler_source is None:
            raise ValueError("Scheduler source is not initialized")

        _logger.info("Starting scheduler")
        await self._result_backend.startup()
        await self._stream_broker.startup()
        await self._scheduler.startup()
        _logger.info("Scheduler started")
        _logger.info("Scheduling task")
        schedules: list[ScheduledTask] = await self._scheduler_source.get_schedules()
        _logger.info("Schedules retrieved", schedules=schedules)

        self._schedule_cron = next(filter(lambda x: x.task_name == "heartbeat", schedules), None)

        if self._schedule_cron is None:
            _logger.info("No schedules found, scheduling task")
            self._dyn_task = self.get_task("heartbeat")
            task_created: CreatedSchedule[Any] = await self._dyn_task.schedule_by_cron(
                source=self._scheduler_source, cron="* * * * *", msg="every minute"
            )
            self._schedule_cron = task_created.task
            _logger.info("Task scheduled")
        else:
            _logger.info("Schedules found, skipping scheduling")

        _logger.info("Starting worker and scheduler tasks")
        taskiq_fastapi.populate_dependency_context(self._stream_broker, app, app.state)  # type: ignore
        self._worker_task: asyncio.Task[None] = asyncio.create_task(run_receiver_task(self._stream_broker))
        self._scheduler_task: asyncio.Task[None] = asyncio.create_task(run_scheduler_task(self._scheduler))
        _logger.info("Worker and scheduler tasks started")

    async def shutdown(self) -> None:
        """Stop the scheduler."""
        _logger.info("Stopping worker")
        self._worker_task.cancel()
        self._scheduler_task.cancel()
        try:
            await self._worker_task
        except (asyncio.CancelledError, RuntimeError) as e:
            _logger.info("Worker task cancelled", error=e)
        try:
            await self._scheduler_task
        except (asyncio.CancelledError, RuntimeError) as e:
            _logger.info("Scheduler task cancelled", error=e)

        while not self._worker_task.done() or not self._scheduler_task.done():
            await asyncio.sleep(0.1)

        _logger.info("Stopping scheduler")
        if self._scheduler is not None:
            await self._scheduler.shutdown()
        if self._stream_broker is not None:
            await self._stream_broker.shutdown()
        if self._result_backend is not None:
            await self._result_backend.shutdown()
        _logger.info("Scheduler stopped")

    @property
    def scheduler(self) -> TaskiqScheduler:
        """Get the scheduler."""
        return cast(TaskiqScheduler, self._scheduler)

    @property
    def broker(self) -> AsyncBroker:
        """Get the broker."""
        return cast(AsyncBroker, self._stream_broker)

    @property
    def scheduler_source(self) -> ScheduleSource:
        """Get the scheduler source."""
        return cast(ScheduleSource, self._scheduler_source)



"""Provides the Book entity."""

from uuid import UUID, uuid4

from pydantic import BaseModel, Field

from .enums import BookType
from .types import BookName


class BookEntity(BaseModel):
    """Book entity."""

    id: UUID = Field(title="Unique identifier of the book", default_factory=uuid4)
    title: BookName = Field(title="Title of the book")
    book_type: BookType = Field(title="Type of book")



"""Provides the status enums for the service."""

from enum import StrEnum


class HealthStatusEnum(StrEnum):
    """Health status enum."""

    HEALTHY = "healthy"
    UNHEALTHY = "unhealthy"


class ReadinessStatusEnum(StrEnum):
    """Readiness status enum."""

    READY = "ready"
    NOT_READY = "not_ready"


class ComponentTypeEnum(StrEnum):
    """Component type enum."""

    SERVICE = "service"
    DATABASE = "database"
    CACHE = "cache"
    STORAGE = "storage"
    MESSAGE_BROKER = "message_broker"
    SEARCH_ENGINE = "search_engine"
    TASK_QUEUE = "task_queue"
    OTHER = "other"



"""Taskiq Plugin Module."""

from importlib.util import find_spec

from .depends import depends_scheduler_component
from .exceptions import TaskiqPluginBaseError
from .plugin import TaskiqPlugin
from .schedulers import SchedulerComponent

__all__: list[str] = [  # pylint: disable=invalid-name
    "SchedulerComponent",
    "TaskiqPlugin",
    "TaskiqPluginBaseError",
    "depends_scheduler_component",
]

if find_spec("beanie") is not None:
    from .depends import depends_odm_database

    __all__ += [  # pylint: disable=invalid-name
        "depends_odm_database",
    ]

if find_spec("aio_pika") is not None:
    from .depends import depends_aiopika_robust_connection

    __all__ += [  # pylint: disable=invalid-name
        "depends_aiopika_robust_connection",
    ]



"""Provides utilities to handle configurations."""

from typing import Any, TypeVar

from pydantic import BaseModel

from fastapi_factory_utilities.core.utils.importlib import get_path_file_in_package
from fastapi_factory_utilities.core.utils.yaml_reader import (
    UnableToReadYamlFileError,
    YamlFileReader,
)

GenericConfigBaseModelType = TypeVar("GenericConfigBaseModelType", bound=BaseModel)  # pylint: disable=invalid-name


class ConfigBaseException(BaseException):
    """Base exception for all the configuration exceptions."""

    pass


class UnableToReadConfigFileError(ConfigBaseException):
    """Exception raised when the configuration file cannot be read.

    Mainly used when the file is not found or the file is not a YAML file.
    """

    pass


class ValueErrorConfigError(ConfigBaseException):
    """Exception raised when the configuration object cannot be created.

    Mainly used when validation fails when creating the configuration object.
    """

    pass


def build_config_from_file_in_package(
    package_name: str,
    filename: str,
    config_class: type[GenericConfigBaseModelType],
    yaml_base_key: str | None = None,
) -> GenericConfigBaseModelType:
    """Build a configuration object from a file in a package.

    Args:
        package_name (str): The package name.
        filename (str): The filename.
        config_class (type[GenericConfigBaseModelType]): The configuration class.
        yaml_base_key (str): The base key in the YAML file.

    Returns:
        GenericConfigBaseModelType: The configuration object.

    Raises:
        UnableToReadConfigFileError: If the configuration file cannot be read.
        ValueErrorConfigError: If the configuration file is invalid.
    """
    # Read the application configuration file
    try:
        yaml_file_content: dict[str, Any] = YamlFileReader(
            file_path=get_path_file_in_package(
                filename=filename,
                package=package_name,
            ),
            yaml_base_key=yaml_base_key,
            use_environment_injection=True,
        ).read()
    except (FileNotFoundError, ImportError, UnableToReadYamlFileError) as exception:
        raise UnableToReadConfigFileError("Unable to read the application configuration file.") from exception

    # Create the application configuration model
    try:
        config: GenericConfigBaseModelType = config_class(**yaml_file_content)
    except ValueError as exception:
        raise ValueErrorConfigError("Unable to create the configuration model.") from exception

    return config



"""Provides the concrete application class."""

from typing import ClassVar

from beanie import Document

from fastapi_factory_utilities.core.app.application import ApplicationAbstract
from fastapi_factory_utilities.core.app.builder import ApplicationGenericBuilder
from fastapi_factory_utilities.core.app.config import RootConfig
from fastapi_factory_utilities.core.plugins.abstracts import PluginAbstract
from fastapi_factory_utilities.core.plugins.odm_plugin import ODMPlugin
from fastapi_factory_utilities.core.plugins.opentelemetry_plugin import OpenTelemetryPlugin
from fastapi_factory_utilities.example.models.books.document import BookDocument


class AppRootConfig(RootConfig):
    """Application configuration class."""

    pass


class App(ApplicationAbstract):
    """Concrete application class."""

    CONFIG_CLASS: ClassVar[type[RootConfig]] = AppRootConfig

    PACKAGE_NAME: ClassVar[str] = "fastapi_factory_utilities.example"

    ODM_DOCUMENT_MODELS: ClassVar[list[type[Document]]] = [BookDocument]

    def configure(self) -> None:
        """Configure the application."""
        # Prevent circular import
        # pylint: disable=import-outside-toplevel
        from .api import api_router  # noqa: PLC0415

        self.get_asgi_app().include_router(router=api_router)

    async def on_startup(self) -> None:
        """Actions to perform on application startup."""
        pass

    async def on_shutdown(self) -> None:
        """Actions to perform on application shutdown."""
        pass


class AppBuilder(ApplicationGenericBuilder[App]):
    """Application builder for the App application."""

    def get_default_plugins(self) -> list[PluginAbstract]:
        """Get the default plugins."""
        return [ODMPlugin(), OpenTelemetryPlugin()]

    def __init__(self, plugins: list[PluginAbstract] | None = None) -> None:
        """Initialize the AppBuilder."""
        # If no plugins are provided, use the default plugins
        if plugins is None:
            plugins = self.get_default_plugins()
        super().__init__(plugins=plugins)



"""Provides utilities for the application."""

import os
from typing import Any

import uvicorn
import uvicorn.server

from fastapi_factory_utilities.core.protocols import ApplicationAbstractProtocol
from fastapi_factory_utilities.core.utils.log import clean_uvicorn_logger


class UvicornUtils:
    """Provides utilities for Uvicorn."""

    def __init__(self, app: ApplicationAbstractProtocol) -> None:
        """Instantiate the factory.

        Args:
            app (BaseApplication): The application.

        Returns:
            None
        """
        self._app: ApplicationAbstractProtocol = app
        self._ssl_keyfile: str | os.PathLike[str] | None = None
        self._ssl_certfile: str | os.PathLike[str] | None = None
        self._ssl_keyfile_password: str | None = None

    def add_ssl_certificates(
        self,
        ssl_keyfile: str | os.PathLike[str] | None = None,
        ssl_certfile: str | os.PathLike[str] | None = None,
        ssl_keyfile_password: str | None = None,
    ) -> None:
        """Add SSL certificates to the application.

        Args:
            ssl_keyfile (str | os.PathLike[str] | None): The SSL key file.
            ssl_certfile (str | os.PathLike[str] | None): The SSL certificate file.
            ssl_keyfile_password (str | None): The SSL key file password.

        Returns:
            None
        """
        self._ssl_keyfile = ssl_keyfile
        self._ssl_certfile = ssl_certfile
        self._ssl_keyfile_password = ssl_keyfile_password

    def build_uvicorn_config(self) -> uvicorn.Config:
        """Build the Uvicorn configuration.

        Returns:
            uvicorn.Config: The Uvicorn configuration.
        """
        kwargs: dict[str, Any] = {}

        if self._ssl_keyfile:
            kwargs["ssl_keyfile"] = self._ssl_keyfile
        if self._ssl_certfile:
            kwargs["ssl_certfile"] = self._ssl_certfile
        if self._ssl_keyfile_password:
            kwargs["ssl_keyfile_password"] = self._ssl_keyfile_password

        config = uvicorn.Config(
            app=self._app.get_asgi_app(),
            host=self._app.get_config().server.host,
            port=self._app.get_config().server.port,
            reload=self._app.get_config().development.reload,
            workers=self._app.get_config().server.workers,
            **kwargs,
        )
        clean_uvicorn_logger()
        return config

    def serve(self) -> None:
        """Serve the application."""
        config: uvicorn.Config = self.build_uvicorn_config()
        server: uvicorn.Server = uvicorn.Server(config=config)
        server.run()



"""Provide FastAPI dependency for ODM."""

from typing import Any

from fastapi import Request
from motor.motor_asyncio import AsyncIOMotorClient, AsyncIOMotorDatabase


def depends_odm_client(request: Request) -> AsyncIOMotorClient[Any]:
    """Acquire the ODM client from the request.

    Args:
        request (Request): The request.

    Returns:
        AsyncIOMotorClient: The ODM client.
    """
    return request.app.state.odm_client


def depends_odm_database(request: Request) -> AsyncIOMotorDatabase[Any]:
    """Acquire the ODM database from the request.

    Args:
        request (Request): The request.

    Returns:
        AsyncIOMotorClient: The ODM database.
    """
    return request.app.state.odm_database



"""Provides a factory function to build a objets for OpenTelemetry."""

from typing import Any, Self
from urllib.parse import ParseResult, urlparse

from opentelemetry.exporter.otlp.proto.grpc.metric_exporter import (
    OTLPMetricExporter as OTLPMetricExporterGRPC,
)
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import (
    OTLPSpanExporter as OTLPSpanExporterGRPC,
)
from opentelemetry.exporter.otlp.proto.http.metric_exporter import (
    OTLPMetricExporter as OTLPMetricExporterHTTP,
)
from opentelemetry.exporter.otlp.proto.http.trace_exporter import (
    OTLPSpanExporter as OTLPSpanExporterHTTP,
)
from opentelemetry.metrics import set_meter_provider
from opentelemetry.propagate import set_global_textmap
from opentelemetry.propagators.b3 import B3MultiFormat
from opentelemetry.sdk.metrics import MeterProvider
from opentelemetry.sdk.metrics.export import MetricReader, PeriodicExportingMetricReader
from opentelemetry.sdk.resources import (
    DEPLOYMENT_ENVIRONMENT,
    SERVICE_NAME,
    SERVICE_NAMESPACE,
    SERVICE_VERSION,
    Resource,
)
from opentelemetry.sdk.trace import SynchronousMultiSpanProcessor, TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor
from opentelemetry.trace import set_tracer_provider

from fastapi_factory_utilities.core.plugins.opentelemetry_plugin.configs import (
    OpenTelemetryMeterConfig,
    OpenTelemetryTracerConfig,
)
from fastapi_factory_utilities.core.protocols import ApplicationAbstractProtocol
from fastapi_factory_utilities.core.utils.importlib import get_path_file_in_package
from fastapi_factory_utilities.core.utils.yaml_reader import (
    UnableToReadYamlFileError,
    YamlFileReader,
)

from .configs import OpenTelemetryConfig, ProtocolEnum
from .exceptions import OpenTelemetryPluginConfigError

GRPC_PORT: int = 4317
HTTP_PORT: int = 4318


class OpenTelemetryPluginBuilder:
    """Configure the injection bindings for OpenTelemetryPlugin."""

    def __init__(self, application: ApplicationAbstractProtocol, settings: OpenTelemetryConfig | None = None) -> None:
        """Instantiate the OpenTelemetryPluginFactory.

        Args:
            application (BaseApplicationProtocol): The application object.
            settings (OpenTelemetryConfig | None): The OpenTelemetry configuration object.
        """
        self._application: ApplicationAbstractProtocol = application
        self._resource: Resource | None = None
        self._config: OpenTelemetryConfig | None = settings
        self._meter_provider: MeterProvider | None = None
        self._tracer_provider: TracerProvider | None = None
        self._trace_exporter: OTLPSpanExporterGRPC | OTLPSpanExporterHTTP | None = None
        self._metric_exporter: OTLPMetricExporterGRPC | OTLPMetricExporterHTTP | None = None

    @property
    def resource(self) -> Resource | None:
        """Provide the resource object for OpenTelemetry.

        Returns:
            Resource | None: The resource object for OpenTelemetry.
        """
        return self._resource

    @property
    def config(self) -> OpenTelemetryConfig | None:
        """Provide the configuration object for OpenTelemetry.

        Returns:
            OpenTelemetryConfig | None: The configuration object for OpenTelemetry.
        """
        return self._config

    @property
    def meter_provider(self) -> MeterProvider | None:
        """Provide the meter provider object for OpenTelemetry.

        Returns:
            MeterProvider | None: The meter provider object for OpenTelemetry.
        """
        return self._meter_provider

    @property
    def tracer_provider(self) -> TracerProvider | None:
        """Provide the tracer provider object for OpenTelemetry.

        Returns:
            TracerProvider | None: The tracer provider object for OpenTelemetry.
        """
        return self._tracer_provider

    def build_resource(self) -> Self:
        """Build a resource object for OpenTelemetry from the application and configs.

        Returns:
            Self: The OpenTelemetryPluginFactory object.
        """
        self._resource = Resource(
            attributes={
                DEPLOYMENT_ENVIRONMENT: self._application.get_config().application.environment.value,
                SERVICE_NAME: self._application.get_config().application.service_name,
                SERVICE_NAMESPACE: self._application.get_config().application.service_namespace,
                SERVICE_VERSION: self._application.get_config().application.version,
            }
        )
        return self

    def build_config(
        self,
    ) -> Self:
        """Build the configuration object for OpenTelemetry from the application.

        Returns:
            Self: The OpenTelemetryPluginFactory object.

        Raises:
            OpenTelemetryPluginConfigError: If the package name is not set in the application.
            OpenTelemetryPluginConfigError: If the application configuration file is not found.

        """
        if self._application.PACKAGE_NAME == "":
            raise OpenTelemetryPluginConfigError("The package name must be set in the concrete application class.")

        # Read the application configuration file
        try:
            yaml_file_content: dict[str, Any] = YamlFileReader(
                file_path=get_path_file_in_package(
                    filename="application.yaml",
                    package=self._application.PACKAGE_NAME,
                ),
                yaml_base_key="opentelemetry",
                use_environment_injection=True,
            ).read()
        except (FileNotFoundError, ImportError, UnableToReadYamlFileError) as exception:
            raise OpenTelemetryPluginConfigError("Unable to read the application configuration file.") from exception

        # Create the application configuration model
        try:
            self._config = OpenTelemetryConfig(**yaml_file_content)
        except ValueError as exception:
            raise OpenTelemetryPluginConfigError("Unable to create the application configuration model.") from exception

        return self

    def build_meter_provider(
        self,
    ) -> Self:
        """Build a meter provider object for OpenTelemetry.

        Returns:
            Self: The OpenTelemetryPluginFactory object.
        """
        if self._resource is None:
            raise OpenTelemetryPluginConfigError("The resource object is missing.")

        if self._config is None:
            raise OpenTelemetryPluginConfigError("The configuration object is missing.")

        readers: list[MetricReader] = []
        if self._config.activate is True:
            if self._config.meter_config is None:
                # TODO: switch to a custom exception
                raise OpenTelemetryPluginConfigError("The meter configuration is missing.")

            # TODO: Extract to a dedicated method for the exporter and period reader setup

            url_parsed: ParseResult = urlparse(self._config.endpoint.unicode_string())

            exporter: OTLPMetricExporterGRPC | OTLPMetricExporterHTTP
            # Setup the Exporter
            if url_parsed.port == GRPC_PORT or self._config.protocol == ProtocolEnum.OTLP_GRPC:
                exporter = OTLPMetricExporterGRPC(
                    endpoint=f"{self._config.endpoint.unicode_string()}",
                    timeout=self._config.timeout,
                    insecure=True if str(self._config.endpoint).startswith("http") else False,
                )
            elif url_parsed.port == HTTP_PORT or self._config.protocol == ProtocolEnum.OTLP_HTTP:
                exporter = OTLPMetricExporterHTTP(
                    endpoint=f"{self._config.endpoint.unicode_string()}/v1/metrics",
                    timeout=self._config.timeout,
                )
            else:
                raise OpenTelemetryPluginConfigError(
                    "The endpoint port is not supported. Use 4317 for gRPC or 4318 for HTTP or set the protocol."
                )

            self._metric_exporter = exporter

            # Setup the Metric Reader
            meter_config: OpenTelemetryMeterConfig = self._config.meter_config
            reader = PeriodicExportingMetricReader(
                exporter=exporter,
                export_interval_millis=meter_config.reader_interval_millis,
                export_timeout_millis=meter_config.reader_timeout_millis,
            )
            readers.append(reader)

        # Setup the Meter Provider
        self._meter_provider = MeterProvider(
            resource=self._resource,
            metric_readers=readers,
            shutdown_on_exit=True,
            views=[],
        )

        set_meter_provider(self._meter_provider)

        return self

    def build_tracer_provider(
        self,
    ) -> Self:
        """Provides a tracer provider for OpenTelemetry.

        Returns:
            Self: The OpenTelemetryPluginFactory object.

        Raises:
            OpenTelemetryPluginConfigError: If the resource object is missing.
            OpenTelemetryPluginConfigError: If the configuration object is missing.
        """
        if self._resource is None:
            raise OpenTelemetryPluginConfigError("The resource object is missing.")

        if self._config is None:
            raise OpenTelemetryPluginConfigError("The configuration object is missing.")

        active_span_processor: SynchronousMultiSpanProcessor | None = None

        # Exit with a void TracerProvider if the export is not activated
        if self._config.activate is True:
            if self._config.tracer_config is None:
                raise OpenTelemetryPluginConfigError("The tracer configuration is missing.")

            exporter: OTLPSpanExporterGRPC | OTLPSpanExporterHTTP
            url_parsed: ParseResult = urlparse(self._config.endpoint.unicode_string())
            # Setup the Exporter
            if url_parsed.port == GRPC_PORT or self._config.protocol == ProtocolEnum.OTLP_GRPC:
                insecure: bool = False if str(self._config.endpoint).startswith("https") else True
                endpoint: str = f"{self._config.endpoint.unicode_string()}"
                exporter = OTLPSpanExporterGRPC(
                    endpoint=endpoint,
                    # timeout=self._config.timeout,
                    insecure=insecure,
                )
            elif url_parsed.port == HTTP_PORT or self._config.protocol == ProtocolEnum.OTLP_HTTP:
                exporter = OTLPSpanExporterHTTP(
                    endpoint=f"{self._config.endpoint.unicode_string()}/v1/traces",
                    # timeout=self._config.timeout,
                )
            else:
                raise OpenTelemetryPluginConfigError(
                    "The endpoint port is not supported. Use 4317 for gRPC or 4318 for HTTP or set the protocol."
                )

            self._trace_exporter = exporter

            # Setup the Span Processor
            tracer_config: OpenTelemetryTracerConfig = self._config.tracer_config
            span_processor = BatchSpanProcessor(
                span_exporter=exporter,
                max_queue_size=tracer_config.max_queue_size,
                max_export_batch_size=tracer_config.max_export_batch_size,
                schedule_delay_millis=tracer_config.schedule_delay_millis,
                export_timeout_millis=tracer_config.export_timeout_millis,
            )

            # Setup the Multi Span Processor
            active_span_processor = SynchronousMultiSpanProcessor()
            active_span_processor.add_span_processor(span_processor=span_processor)

            # Setup the TextMap Propagator for B3
            set_global_textmap(http_text_format=B3MultiFormat())

        # Setup the Tracer Provider
        self._tracer_provider = TracerProvider(
            sampler=None,
            resource=self._resource,
            active_span_processor=active_span_processor,
            id_generator=None,
            span_limits=None,
            shutdown_on_exit=True,
        )

        set_tracer_provider(self._tracer_provider)

        return self

    def build_all(
        self,
    ) -> Self:
        """Build all the objects for OpenTelemetry.

        Returns:
            Self: The OpenTelemetryPluginFactory object.
        """
        self.build_resource()
        if self._config is None:
            self.build_config()
        self.build_meter_provider()
        self.build_tracer_provider()

        return self



"""Package covering the StatusService."""

from .enums import ComponentTypeEnum, HealthStatusEnum, ReadinessStatusEnum
from .services import StatusService
from .types import ComponentInstanceType, Status

__all__: list[str] = [
    "ComponentInstanceType",
    "ComponentTypeEnum",
    "HealthStatusEnum",
    "ReadinessStatusEnum",
    "Status",
    "StatusService",
]



"""Provide the API for the example package."""

from fastapi import APIRouter

from .books import api_v1_books_router, api_v2_books_router

api_router: APIRouter = APIRouter(prefix="/api")

# -- API v1
api_router_v1: APIRouter = APIRouter(prefix="/v1")
api_router_v1.include_router(router=api_v1_books_router)

# -- API v2
api_router_v2: APIRouter = APIRouter(prefix="/v2")
api_router_v2.include_router(router=api_v2_books_router)

# -- Include API versions
api_router.include_router(router=api_router_v1)
api_router.include_router(router=api_router_v2)



"""Provides the abstract class for the exchange port for the Aiopika plugin."""

from typing import ClassVar, Self

from aio_pika import Exchange as AiopikaExchange
from aio_pika import ExchangeType
from aio_pika.abc import TimeoutType

from .abstract import AbstractAiopikaResource
from .exceptions import AiopikaPluginBaseError, AiopikaPluginExchangeNotDeclaredError


class Exchange(AbstractAiopikaResource):
    """Abstract class for the exchange port for the Aiopika plugin."""

    DEFAULT_OPERATION_TIMEOUT: ClassVar[TimeoutType] = 10.0

    def __init__(  # pylint: disable=too-many-arguments # noqa: PLR0913
        self,
        name: str,
        exchange_type: ExchangeType,
        durable: bool = True,
        auto_delete: bool = False,
        internal: bool = False,
        passive: bool = False,
        timeout: TimeoutType = DEFAULT_OPERATION_TIMEOUT,
    ) -> None:
        """Initialize the exchange port."""
        super().__init__()
        self._name: str = name
        self._exchange_type: ExchangeType = exchange_type
        self._durable: bool = durable
        self._auto_delete: bool = auto_delete
        self._internal: bool = internal
        self._passive: bool = passive
        self._timeout: TimeoutType = timeout
        self._aiopika_exchange: AiopikaExchange | None = None
        self._is_declared: bool = False

    @property
    def exchange(self) -> AiopikaExchange:
        """Get the Aiopika exchange."""
        if self._aiopika_exchange is None:
            raise AiopikaPluginExchangeNotDeclaredError(message="Exchange not declared.", exchange=self._name)
        return self._aiopika_exchange

    async def _declare(self) -> Self:
        """Declare the exchange."""
        try:
            self._aiopika_exchange = await self._channel.declare_exchange(  # pyright: ignore
                name=self._name,
                type=self._exchange_type,
                durable=self._durable,
                auto_delete=self._auto_delete,
                internal=self._internal,
                passive=self._passive,
                timeout=self._timeout,
            )
        except Exception as exception:
            raise AiopikaPluginBaseError(
                message="Failed to declare the exchange.",
            ) from exception
        return self

    async def setup(self) -> Self:
        """Setup the exchange."""
        await super().setup()
        if self._aiopika_exchange is None:
            await self._declare()
        return self



"""API v1 sys health module.

Provide the Get health endpoint
"""

from http import HTTPStatus

from fastapi import APIRouter, Depends, Response
from pydantic import BaseModel

from fastapi_factory_utilities.core.services.status.enums import HealthStatusEnum
from fastapi_factory_utilities.core.services.status.services import (
    StatusService,
    depends_status_service,
)
from fastapi_factory_utilities.core.services.status.types import ComponentInstanceKey

api_v1_sys_health = APIRouter(prefix="/health")


class HealthResponseModel(BaseModel):
    """Health response schema."""

    status: HealthStatusEnum


@api_v1_sys_health.get(
    path="",
    tags=["sys"],
    response_model=HealthResponseModel,
    responses={
        HTTPStatus.OK.value: {
            "model": HealthResponseModel,
            "description": "Health status.",
        },
        HTTPStatus.INTERNAL_SERVER_ERROR.value: {
            "model": HealthResponseModel,
            "description": "Internal server error.",
        },
    },
)
def get_api_v1_sys_health(
    response: Response, status_service: StatusService = Depends(depends_status_service)
) -> HealthResponseModel:
    """Get the health of the system.

    Args:
        response (Response): The response object.
        status_service (StatusService): The status service.

    Returns:
        HealthResponse: The health status.
    """
    status: HealthStatusEnum = status_service.get_status()["health"]
    match status:
        case HealthStatusEnum.HEALTHY:
            response.status_code = HTTPStatus.OK.value
        case HealthStatusEnum.UNHEALTHY:
            response.status_code = HTTPStatus.INTERNAL_SERVER_ERROR.value
    return HealthResponseModel(status=status)


class ComponentHealthResponseModel(BaseModel):
    """Component health response schema."""

    components: dict[ComponentInstanceKey, HealthStatusEnum]


@api_v1_sys_health.get(
    path="/components",
    tags=["sys"],
    response_model=ComponentHealthResponseModel,
    responses={
        HTTPStatus.OK.value: {
            "model": ComponentHealthResponseModel,
            "description": "Health status of all components.",
        },
    },
)
def get_api_v1_sys_components_health(
    status_service: StatusService = Depends(depends_status_service),
) -> ComponentHealthResponseModel:
    """Get the health of all components.

    Args:
        status_service (StatusService): The status service.

    Returns:
        list[ComponentHealthResponseModel]: The health status of all components.
    """
    components_dict: dict[ComponentInstanceKey, HealthStatusEnum] = {}

    for _, components in status_service.get_components_status_by_type().items():
        for key, status in components.items():
            components_dict[key] = status["health"]

    return ComponentHealthResponseModel(components=dict(components_dict))



"""Provides the exceptions for the ODM_Plugin."""


class ODMPluginBaseException(BaseException):
    """Base exception for the ODM_Plugin."""

    pass


class ODMPluginConfigError(ODMPluginBaseException):
    """Exception for the ODM_Plugin configuration."""

    pass


class UnableToCreateEntityDueToDuplicateKeyError(ODMPluginBaseException):
    """Exception for when the entity cannot be created due to a duplicate key error."""

    pass


class OperationError(ODMPluginBaseException):
    """Exception for when an operation fails."""

    pass



"""Provides the Kratos Objects."""

import datetime
import uuid
from typing import ClassVar

from pydantic import BaseModel, ConfigDict

from .enums import AuthenticatorAssuranceLevelEnum


class KratosTraitsObject(BaseModel):
    """Traits for Kratos."""

    model_config: ClassVar[ConfigDict] = ConfigDict(extra="ignore")

    email: str
    username: str
    realm_id: uuid.UUID


class KratosIdentityObject(BaseModel):
    """Identity for Kratos."""

    model_config: ClassVar[ConfigDict] = ConfigDict(extra="ignore")

    id: uuid.UUID
    state: str
    traits: KratosTraitsObject


class KratosSessionObject(BaseModel):
    """Session object for Kratos."""

    model_config: ClassVar[ConfigDict] = ConfigDict(extra="ignore")

    id: uuid.UUID
    active: bool
    issued_at: datetime.datetime
    expires_at: datetime.datetime
    authenticated_at: datetime.datetime
    authenticator_assurance_level: AuthenticatorAssuranceLevelEnum
    identity: KratosIdentityObject



"""Provides the OpenTelemetry plugin."""

import asyncio
from typing import Self, cast

from fastapi import Request
from opentelemetry.sdk.metrics import MeterProvider
from opentelemetry.sdk.trace import TracerProvider
from structlog.stdlib import BoundLogger, get_logger

from fastapi_factory_utilities.core.plugins.abstracts import PluginAbstract

from .builder import OpenTelemetryPluginBuilder
from .configs import OpenTelemetryConfig
from .exceptions import OpenTelemetryPluginBaseException, OpenTelemetryPluginConfigError
from .instruments import INSTRUMENTS

__all__: list[str] = [
    "OpenTelemetryConfig",
    "OpenTelemetryPluginBaseException",
    "OpenTelemetryPluginBuilder",
    "OpenTelemetryPluginConfigError",
]

_logger: BoundLogger = get_logger()


class OpenTelemetryPlugin(PluginAbstract):
    """OpenTelemetry plugin."""

    SECONDS_TO_MS_MULTIPLIER: int = 1000

    def __init__(self) -> None:
        """Initialize the OpenTelemetry plugin."""
        super().__init__()
        self._otel_config: OpenTelemetryConfig | None = None
        self._tracer_provider: TracerProvider | None = None
        self._meter_provider: MeterProvider | None = None

    def _build(self) -> Self:
        """Build the OpenTelemetry plugin."""
        assert self._application is not None
        # Build the OpenTelemetry Resources, TracerProvider and MeterProvider
        try:
            otel_builder: OpenTelemetryPluginBuilder = OpenTelemetryPluginBuilder(
                application=self._application
            ).build_all()
        except OpenTelemetryPluginBaseException as exception:
            _logger.error(f"OpenTelemetry plugin failed to start. {exception}")
            raise
        # Configuration is never None at this point (checked in the builder and raises an exception)
        self._otel_config = cast(OpenTelemetryConfig, otel_builder.config)
        self._tracer_provider = otel_builder.tracer_provider
        self._meter_provider = otel_builder.meter_provider
        return self

    def _instrument(self) -> None:
        """Instrument the FastAPI application."""
        assert self._application is not None
        assert self._tracer_provider is not None
        assert self._meter_provider is not None
        assert self._otel_config is not None

        for instrument in INSTRUMENTS:
            instrument(self._application, self._otel_config, self._meter_provider, self._tracer_provider)

    def on_load(self) -> None:
        """On load."""
        assert self._application is not None
        # Build the OpenTelemetry Resources, TracerProvider and MeterProvider
        self._build()
        assert self._tracer_provider is not None
        assert self._meter_provider is not None
        assert self._otel_config is not None
        self._add_to_state(key="tracer_provider", value=self._tracer_provider)
        self._add_to_state(key="meter_provider", value=self._meter_provider)
        self._add_to_state(key="otel_config", value=self._otel_config)
        # Instrument the FastAPI application and AioHttpClient, ...
        self._instrument()
        _logger.debug(f"OpenTelemetry plugin loaded. {self._otel_config.activate=}")

    async def on_startup(self) -> None:
        """On startup."""
        pass

    async def close_tracer_provider(self) -> None:
        """Close the tracer provider."""
        assert self._tracer_provider is not None
        assert self._otel_config is not None
        self._tracer_provider.force_flush(
            timeout_millis=self._otel_config.closing_timeout * self.SECONDS_TO_MS_MULTIPLIER
        )
        # No Delay for the shutdown of the tracer provider
        try:
            self._tracer_provider.shutdown()
        except Exception as exception:  # pylint: disable=broad-exception-caught
            _logger.error("OpenTelemetry plugin failed to close the tracer provider.", error=exception)

    async def close_meter_provider(self) -> None:
        """Close the meter provider."""
        assert self._meter_provider is not None
        assert self._otel_config is not None
        self._meter_provider.force_flush(
            timeout_millis=self._otel_config.closing_timeout * self.SECONDS_TO_MS_MULTIPLIER
        )
        try:
            self._meter_provider.shutdown(
                timeout_millis=self._otel_config.closing_timeout * self.SECONDS_TO_MS_MULTIPLIER
            )
        except Exception as exception:  # pylint: disable=broad-exception-caught
            _logger.error("OpenTelemetry plugin failed to close the meter provider.", error=exception)

    async def on_shutdown(self) -> None:
        """On shutdown."""
        _logger.debug("OpenTelemetry plugin stop requested. Flushing and closing...")

        await asyncio.gather(
            self.close_tracer_provider(),
            self.close_meter_provider(),
        )

        _logger.debug("OpenTelemetry plugin closed.")


def depends_tracer_provider(request: Request) -> TracerProvider:
    """Get the tracer provider."""
    return request.app.state.tracer_provider


def depends_meter_provider(request: Request) -> MeterProvider:
    """Get the meter provider."""
    return request.app.state.meter_provider


def depends_otel_config(request: Request) -> OpenTelemetryConfig:
    """Get the OpenTelemetry config."""
    return request.app.state.otel_config



"""Provides the core application module for the Python Factory."""

from .application import ApplicationAbstract
from .builder import ApplicationGenericBuilder
from .config import (
    BaseApplicationConfig,
    DependencyConfig,
    HttpServiceDependencyConfig,
    RootConfig,
    depends_dependency_config,
)
from .enums import EnvironmentEnum

__all__: list[str] = [
    "ApplicationAbstract",
    "ApplicationGenericBuilder",
    "BaseApplicationConfig",
    "DependencyConfig",
    "EnvironmentEnum",
    "HttpServiceDependencyConfig",
    "RootConfig",
    "depends_dependency_config",
]



"""Provides the response objects for the books API."""

from uuid import UUID

from pydantic import BaseModel, ConfigDict

from fastapi_factory_utilities.example.entities.books import BookName, BookType


class BookResponseModel(BaseModel):
    """Book response model."""

    model_config = ConfigDict(extra="ignore")

    id: UUID
    title: BookName
    book_type: BookType


class BookListReponse(BaseModel):
    """Book list response."""

    model_config = ConfigDict(extra="forbid")

    books: list[BookResponseModel]
    size: int



"""Provides the Aiopika plugin."""

from typing import cast

from aio_pika import connect_robust  # pyright: ignore[reportUnknownMemberType]
from aio_pika.abc import AbstractRobustConnection
from fastapi import Request
from opentelemetry.instrumentation.aio_pika import AioPikaInstrumentor  # pyright: ignore[reportMissingTypeStubs]
from opentelemetry.sdk.metrics import MeterProvider
from opentelemetry.sdk.trace import TracerProvider
from structlog.stdlib import BoundLogger, get_logger

from fastapi_factory_utilities.core.plugins.abstracts import PluginAbstract

from .configs import AiopikaConfig, build_config_from_package
from .depends import DEPENDS_AIOPIKA_ROBUST_CONNECTION_KEY
from .exceptions import AiopikaPluginBaseError

_logger: BoundLogger = get_logger(__package__)


class AiopikaPlugin(PluginAbstract):
    """Aiopika plugin."""

    def __init__(self, aiopika_config: AiopikaConfig | None = None) -> None:
        """Initialize the Aiopika plugin."""
        super().__init__()
        self._aiopika_config: AiopikaConfig | None = aiopika_config
        self._robust_connection: AbstractRobustConnection | None = None

    @property
    def robust_connection(self) -> AbstractRobustConnection:
        """Get the robust connection."""
        assert self._robust_connection is not None
        return self._robust_connection

    def on_load(self) -> None:
        """On load."""
        assert self._application is not None

        # Build the configuration if not provided
        if self._aiopika_config is None:
            self._aiopika_config = build_config_from_package(package_name=self._application.PACKAGE_NAME)

    async def on_startup(self) -> None:
        """On startup."""
        assert self._application is not None
        assert self._aiopika_config is not None

        tracer_provider: TracerProvider | None = cast(
            TracerProvider | None, getattr(self._application.get_asgi_app().state, "tracer_provider", None)
        )
        meter_provider: MeterProvider | None = cast(
            MeterProvider | None, getattr(self._application.get_asgi_app().state, "meter_provider", None)
        )
        if tracer_provider is None or meter_provider is None:
            raise AiopikaPluginBaseError("Tracer provider or meter provider not found in the application state.")

        AioPikaInstrumentor().instrument(
            tracer_provider=tracer_provider,
            meter_provider=meter_provider,
        )

        self._robust_connection = await connect_robust(url=str(self._aiopika_config.amqp_url))
        self._add_to_state(key=DEPENDS_AIOPIKA_ROBUST_CONNECTION_KEY, value=self._robust_connection)
        _logger.debug("Aiopika plugin connected to the AMQP server.", amqp_url=self._aiopika_config.amqp_url)

    async def on_shutdown(self) -> None:
        """On shutdown."""
        if self._robust_connection is not None:
            await self._robust_connection.close()
        _logger.debug("Aiopika plugin shutdown.")


def depends_robust_connection(request: Request) -> AbstractRobustConnection:
    """Depends on the robust connection.

    Args:
        request (Request): The request.

    Returns:
        AbstractRobustConnection: The robust connection.
    """
    return request.app.state.robust_connection



"""Hydra service module."""

from .exceptions import HydraOperationError, HydraTokenInvalidError
from .objects import HydraTokenIntrospectObject
from .services import HydraService, depends_hydra_service

__all__: list[str] = [
    "HydraOperationError",
    "HydraService",
    "HydraTokenIntrospectObject",
    "HydraTokenInvalidError",
    "depends_hydra_service",
]



"""Provides the Taskiq plugin."""

from collections.abc import Callable

from fastapi_factory_utilities.core.plugins.abstracts import PluginAbstract

from .configs import RedisCredentialsConfig
from .depends import DEPENDS_SCHEDULER_COMPONENT_KEY
from .schedulers import SchedulerComponent


class TaskiqPlugin(PluginAbstract):
    """Taskiq plugin."""

    def __init__(
        self, redis_credentials_config: RedisCredentialsConfig, register_hook: Callable[[SchedulerComponent], None]
    ) -> None:
        """Initialize the Taskiq plugin."""
        super().__init__()
        self._redis_credentials_config: RedisCredentialsConfig = redis_credentials_config
        self._register_hook: Callable[[SchedulerComponent], None] = register_hook
        self._scheduler_component: SchedulerComponent = SchedulerComponent()

    def on_load(self) -> None:
        """On load."""
        assert self._application is not None
        self._scheduler_component.configure(
            redis_connection_string=self._redis_credentials_config.url, app=self._application.get_asgi_app()
        )
        self._add_to_state(key=DEPENDS_SCHEDULER_COMPONENT_KEY, value=self._scheduler_component)
        self._register_hook(self._scheduler_component)

    async def on_startup(self) -> None:
        """On startup."""
        assert self._application is not None
        await self._scheduler_component.startup(app=self._application.get_asgi_app())

    async def on_shutdown(self) -> None:
        """On shutdown."""
        assert self._application is not None
        await self._scheduler_component.shutdown()



"""FastAPI Factory Utilities exceptions."""

import logging
from collections.abc import Sequence
from typing import NotRequired, TypedDict, Unpack

from opentelemetry.trace import Span, get_current_span
from opentelemetry.util.types import AttributeValue
from structlog.stdlib import BoundLogger, get_logger

_logger: BoundLogger = get_logger()


class ExceptionParameters(TypedDict):
    """Parameters for the exception."""

    message: NotRequired[str]
    level: NotRequired[int]


class FastAPIFactoryUtilitiesError(Exception):
    """Base exception for the FastAPI Factory Utilities."""

    def __init__(
        self,
        *args: object,
        **kwargs: Unpack[ExceptionParameters],
    ) -> None:
        """Instantiate the exception.

        Args:
            *args: The arguments.
            message: The message.
            level: The logging level.
            **kwargs: The keyword arguments.

        """
        # Extract the message and the level from the kwargs if they are present
        self.message: str | None = kwargs.pop("message", None)
        self.level: int = kwargs.pop("level", logging.ERROR)

        # If the message is not present, try to extract it from the args
        if self.message is None and len(args) > 0 and isinstance(args[0], str):
            self.message = args[0]

        # Log the Exception
        if self.message:
            _logger.log(level=self.level, event=self.message)

        try:
            # Propagate the exception
            span: Span = get_current_span()
            # If not otel is setup, INVALID_SPAN is retrieved from get_current_span
            # and it will respond False to the is_recording method
            if span.is_recording():
                span.record_exception(self)
                for key, value in kwargs.items():
                    attribute_value: AttributeValue
                    if not isinstance(value, (str, bool, int, float, Sequence)):
                        attribute_value = str(value)
                    else:
                        attribute_value = value
                    span.set_attribute(key, attribute_value)
        except Exception:  # pylint: disable=broad-exception-caught
            # Suppress any errors that occur while propagating the exception
            pass

        # Call the parent class
        super().__init__(*args)

    def __str__(self) -> str:
        """Return the string representation of the exception.

        Returns:
            str: The message if available, otherwise the default exception string.
        """
        if self.message is not None:
            return self.message
        return super().__str__()



"""Repository for books."""

from fastapi_factory_utilities.core.plugins.odm_plugin.repositories import (
    AbstractRepository,
)
from fastapi_factory_utilities.example.entities.books import BookEntity
from fastapi_factory_utilities.example.models.books.document import BookDocument


class BookRepository(AbstractRepository[BookDocument, BookEntity]):
    """Repository for books."""



"""Provides the dependencies for the Taskiq plugin."""

from importlib.util import find_spec
from typing import TYPE_CHECKING, Any

from fastapi import Request
from taskiq import TaskiqDepends

if TYPE_CHECKING:
    from .schedulers import SchedulerComponent

DEPENDS_SCHEDULER_COMPONENT_KEY: str = "scheduler_component"


def depends_scheduler_component(
    request: Request = TaskiqDepends(),
) -> "SchedulerComponent":
    """Dependency injection for the scheduler component."""
    return getattr(request.app.state, DEPENDS_SCHEDULER_COMPONENT_KEY)


if find_spec("beanie") is not None:
    from motor.motor_asyncio import AsyncIOMotorDatabase

    def depends_odm_database(request: Request = TaskiqDepends()) -> AsyncIOMotorDatabase[Any]:
        """Acquire the ODM database from the request.

        Args:
            request (Request): The request.

        Returns:
            AsyncIOMotorClient: The ODM database.
        """
        return request.app.state.odm_database


if find_spec("aio_pika") is not None:
    from aio_pika.abc import AbstractRobustConnection

    from fastapi_factory_utilities.core.plugins.aiopika.depends import DEPENDS_AIOPIKA_ROBUST_CONNECTION_KEY

    def depends_aiopika_robust_connection(request: Request = TaskiqDepends()) -> AbstractRobustConnection:
        """Acquire the Aiopika robust connection from the request.

        Args:
            request (Request): The request.

        Returns:
            AbstractRobustConnection: The Aiopika robust connection.
        """
        return getattr(request.app.state, DEPENDS_AIOPIKA_ROBUST_CONNECTION_KEY)



"""Provides the message for the Aiopika plugin."""

from enum import StrEnum, auto
from typing import ClassVar, Generic, TypeVar

from aio_pika.abc import DeliveryMode, HeadersType
from aio_pika.message import IncomingMessage, Message
from pydantic import BaseModel, ConfigDict, Field, PrivateAttr

GenericMessageData = TypeVar("GenericMessageData", bound=BaseModel)


class SenderModel(BaseModel):
    """Sender model."""

    model_config: ClassVar[ConfigDict] = ConfigDict(extra="forbid", frozen=True)

    name: str = Field(description="The name of the sender.")


class MessageTypeEnum(StrEnum):
    """Message type enum."""

    FUNCTIONAL_EVENT = auto()


class AbstractMessage(BaseModel, Generic[GenericMessageData]):
    """Abstract message."""

    model_config: ClassVar[ConfigDict] = ConfigDict(extra="forbid", frozen=True)

    message_type: MessageTypeEnum = Field(
        description="The type of the message.", default=MessageTypeEnum.FUNCTIONAL_EVENT
    )
    sender: SenderModel = Field(description="The sender of the message.")
    data: GenericMessageData = Field(description="The data of the message.")

    _incoming_message: IncomingMessage | None = PrivateAttr()
    _headers: HeadersType = PrivateAttr(default_factory=dict)

    def get_headers(self) -> HeadersType:
        """Get the headers of the message."""
        return self._headers

    def set_headers(self, headers: HeadersType) -> None:
        """Set the headers of the message."""
        self._headers = headers

    def set_incoming_message(self, incoming_message: IncomingMessage) -> None:
        """Set the incoming message."""
        self._incoming_message = incoming_message
        self.set_headers(headers=incoming_message.headers)

    async def ack(self) -> None:
        """Ack the message.

        Raises:
            - ValueError: If the incoming message is not set.
        """
        if self._incoming_message is None:
            raise ValueError("Incoming message is not set.")
        await self._incoming_message.ack(multiple=False)

    async def reject(self, requeue: bool = True) -> None:
        """Reject the message.

        Args:
            requeue (bool): Whether to requeue the message.

        Raises:
            - ValueError: If the incoming message is not set.
        """
        if self._incoming_message is None:
            raise ValueError("Incoming message is not set.")
        await self._incoming_message.reject(requeue=requeue)

    def to_aiopika_message(self) -> Message:
        """Convert the message to an Aiopika message."""
        return Message(
            body=self.model_dump_json().encode("utf-8"),
            headers=self.get_headers(),
            content_type="application/json",
            content_encoding="utf-8",
            delivery_mode=DeliveryMode.PERSISTENT,
            priority=0,
        )



"""Health calculator strategies."""

from typing import Protocol, runtime_checkable

from .enums import HealthStatusEnum
from .types import ComponentInstanceKey, Status


@runtime_checkable
class HealthCalculatorStrategy(Protocol):
    """Health calculator strategy."""

    def __init__(self, components_status: dict[ComponentInstanceKey, Status]) -> None:
        """Initialize the health calculator.

        Args:
            components_status (dict[str, Status]): The components status.
        """
        raise NotImplementedError

    def calculate(self) -> HealthStatusEnum:
        """Calculate the health status."""
        raise NotImplementedError


class HealthSimpleCalculatorStrategy:
    """Health calculator.

    This class calculates the health status based on the components status.
    It's a simple implementation that returns unhealthy if at least one component is unhealthy for now.
    We want to enhance this class to support more complex health calculation in the future.
    ( )
    """

    def __init__(self, components_status: dict[ComponentInstanceKey, Status]) -> None:
        """Initialize the health calculator.

        Args:
            components_status (dict[str, Status]): The components status.
        """
        self._components_status: dict[ComponentInstanceKey, Status] = components_status

    def calculate(self) -> HealthStatusEnum:
        """Calculate the health status."""
        for status in self._components_status.values():
            if status["health"] != HealthStatusEnum.HEALTHY:
                return HealthStatusEnum.UNHEALTHY
        return HealthStatusEnum.HEALTHY



"""Protocols for the base application."""

from abc import abstractmethod
from typing import TYPE_CHECKING, ClassVar, Protocol

from beanie import Document
from fastapi import FastAPI

from fastapi_factory_utilities.core.services.status.services import StatusService

if TYPE_CHECKING:
    from fastapi_factory_utilities.core.app.config import RootConfig


class ApplicationAbstractProtocol(Protocol):
    """Protocol for the base application."""

    PACKAGE_NAME: ClassVar[str]

    ODM_DOCUMENT_MODELS: ClassVar[list[type[Document]]]

    @abstractmethod
    def get_config(self) -> "RootConfig":
        """Get the application configuration."""

    @abstractmethod
    def get_asgi_app(self) -> FastAPI:
        """Get the ASGI application."""

    @abstractmethod
    def get_status_service(self) -> StatusService:
        """Get the status service."""



"""Provides the KratosService class for handling Kratos operations."""

from http import HTTPStatus
from typing import Annotated

import aiohttp
from fastapi import Depends
from pydantic import ValidationError

from fastapi_factory_utilities.core.app import (
    DependencyConfig,
    HttpServiceDependencyConfig,
    depends_dependency_config,
)

from .exceptions import KratosOperationError, KratosSessionInvalidError
from .objects import KratosSessionObject


class KratosService:
    """Service class for handling Kratos operations."""

    COOKIE_NAME: str = "ory_kratos_session"

    def __init__(self, kratos_http_config: HttpServiceDependencyConfig) -> None:
        """Initialize the KratosService class.

        Args:
            kratos_http_config (HttpServiceDependencyConfig): Kratos HTTP configuration.
        """
        self._http_config: HttpServiceDependencyConfig = kratos_http_config

    async def whoami(self, cookie_value: str) -> KratosSessionObject:
        """Get the current user session.

        Args:
            cookie_value (str): Cookie value.

        Returns:
            KratosSessionObject: Kratos session object.

        Raises:
            KratosOperationError: If the Kratos service returns an error.
            KratosSessionInvalidError: If the Kratos session is invalid.
        """
        cookies: dict[str, str] = {self.COOKIE_NAME: cookie_value}
        async with aiohttp.ClientSession(base_url=str(self._http_config.url), cookies=cookies) as session:
            async with session.get(
                url="/sessions/whoami",
            ) as response:
                if response.status >= HTTPStatus.INTERNAL_SERVER_ERROR.value:
                    raise KratosOperationError(message=f"Kratos service error: {response.status} - {response.reason}")
                if response.status == HTTPStatus.UNAUTHORIZED:
                    raise KratosSessionInvalidError(
                        message=f"Kratos session invalid: {response.status} - {response.reason}"
                    )
                if response.status != HTTPStatus.OK:
                    raise KratosOperationError(message=f"Kratos service error: {response.status} - {response.reason}")

                try:
                    kratos_session: KratosSessionObject = KratosSessionObject(**await response.json())
                except ValidationError as e:
                    raise KratosOperationError(message=f"Kratos service error: {e}") from e

                return kratos_session


def depends_kratos_service(
    dependency_config: Annotated[DependencyConfig, Depends(depends_dependency_config)],
) -> KratosService:
    """Dependency function to get the Kratos service instance.

    Args:
        dependency_config (DependencyConfig): Dependency configuration.

    Returns:
        KratosService: Kratos service instance.

    Raises:
        KratosOperationError: If the Kratos dependency is not configured.
    """
    if dependency_config.kratos is None:
        raise KratosOperationError(message="Kratos dependency not configured")
    return KratosService(
        kratos_http_config=dependency_config.kratos,
    )



"""Provides enums for the books service."""

from enum import StrEnum, auto


class BookType(StrEnum):
    """Enumeration of book types."""

    SCIENCE_FICTION = auto()
    FANTASY = auto()
    MYSTERY = auto()
    ROMANCE = auto()
    THRILLER = auto()
    HORROR = auto()
    HISTORICAL_FICTION = auto()
    ADVENTURE = auto()



"""Provides exceptions for the status service."""

from fastapi_factory_utilities.core.exceptions import FastAPIFactoryUtilitiesError
from fastapi_factory_utilities.core.services.status.types import ComponentInstanceType


class StatusServiceError(FastAPIFactoryUtilitiesError):
    """Status service error."""


class ComponentRegistrationError(StatusServiceError):
    """Component registration error."""

    def __init__(
        self,
        component_instance: ComponentInstanceType,
    ) -> None:
        """Initialize the component registration error.

        Args:
            component_instance (ComponentInstanceType): The component instance.

        """
        super().__init__(
            message="An error occurred while registering the component instance.",
            component_instance=component_instance,  # type: ignore
        )



"""Provides the abstract class for the listener port for the Aiopika plugin."""

from abc import abstractmethod
from collections.abc import Awaitable, Callable
from typing import Any, ClassVar, Generic, Self, TypeVar, cast, get_args

from aio_pika.abc import ConsumerTag, TimeoutType
from aio_pika.message import IncomingMessage

from ..abstract import AbstractAiopikaResource
from ..message import AbstractMessage
from ..queue import Queue

GenericMessage = TypeVar("GenericMessage", bound=AbstractMessage[Any])


class AbstractListener(AbstractAiopikaResource, Generic[GenericMessage]):
    """Abstract class for the listener port for the Aiopika plugin."""

    DEFAULT_OPERATION_TIMEOUT: ClassVar[TimeoutType] = 10.0

    def __init__(self, queue: Queue, name: str | None = None) -> None:
        """Initialize the listener port."""
        super().__init__()
        self._name: str = name or self.__class__.__name__
        self._queue: Queue = queue
        self._consumer_tag: ConsumerTag | None = None
        generic_args: tuple[Any, ...] = get_args(self.__orig_bases__[0])  # type: ignore
        self._message_type: type[GenericMessage] = generic_args[0]

    async def setup(self) -> Self:
        """Setup the listener."""
        await super().setup()
        await self._queue.setup()
        return self

    async def listen(self) -> None:
        """Listen for messages."""
        self._consumer_tag = await self._queue.queue.consume(  # pyright: ignore
            callback=cast(Callable[[IncomingMessage], Awaitable[Any]], self._on_message),  # pyright: ignore
            exclusive=True,
        )

    async def _on_message(self, incoming_message: IncomingMessage) -> None:
        """On message."""
        message: GenericMessage = self._message_type.model_validate_json(incoming_message.body)
        message.set_incoming_message(incoming_message=incoming_message)
        await self.on_message(message=message)

    async def close(self) -> None:
        """Close the listener.

        Returns:
            - None: The listener is closed.

        Raises:
            - AiopikaPluginBaseException: If the listener cannot be closed.
        """
        if self._consumer_tag is not None:
            await self._queue.queue.cancel(consumer_tag=self._consumer_tag)

    @abstractmethod
    async def on_message(self, message: GenericMessage) -> None:
        """On message.

        Args:
            message (GenericMessage): The message.

        Returns:
            - None: The message is processed.
        """
        raise NotImplementedError



"""Aiopika Plugin Module."""

from .configs import AiopikaConfig
from .depends import depends_aiopika_robust_connection
from .exceptions import AiopikaPluginBaseError, AiopikaPluginConfigError
from .exchange import Exchange
from .listener import AbstractListener
from .message import AbstractMessage, SenderModel
from .plugins import AiopikaPlugin
from .publisher import AbstractPublisher
from .queue import Queue

__all__: list[str] = [
    "AbstractListener",
    "AbstractMessage",
    "AbstractPublisher",
    "AiopikaConfig",
    "AiopikaPlugin",
    "AiopikaPluginBaseError",
    "AiopikaPluginConfigError",
    "Exchange",
    "Queue",
    "SenderModel",
    "depends_aiopika_robust_connection",
]



"""Provides the exceptions for the Taskiq plugin."""

from typing import Any

from fastapi_factory_utilities.core.exceptions import FastAPIFactoryUtilitiesError


class TaskiqPluginBaseError(FastAPIFactoryUtilitiesError):
    """Base class for all exceptions raised by the Taskiq plugin."""

    def __init__(self, message: str, **kwargs: Any) -> None:
        """Initialize the Taskiq plugin base exception."""
        super().__init__(message, **kwargs)



"""Provides the monitored abstract class for monitoring the status of the application.

```python
# Example of using the MonitoredAbstract


class MyMonitored(MonitoredAbstract):
    def __init__(self, status_service: StatusService) -> None:
        super().__init__(
            component_instance=ComponentInstanceType(
                component_type=ComponentTypeEnum.APPLICATION,
                component_name="MyMonitored",
            ),
            status_service=status_service,
        )

    def my_method(self) -> None:
        self.update_monitoring_status(
            Status(
                status_type=StatusTypeEnum.INFO,
                status_message="My method is running.",
            )
        )
```

"""

from abc import ABC

from reactivex import Subject

from fastapi_factory_utilities.core.services.status import (
    ComponentInstanceType,
    ComponentTypeEnum,
    Status,
    StatusService,
)


class MonitoredAbstract(ABC):
    """Monitored abstract class."""

    def __init__(self, component_instance: ComponentInstanceType, status_service: StatusService) -> None:
        """Initialize the monitored.

        Args:
            component_instance (ComponentInstanceType): The component instance.
            status_service (StatusService): The status service.

        """
        self._monit_component_instance: ComponentInstanceType = component_instance
        self._monit_status_service_subject: Subject[Status] = status_service.register_component_instance(
            component_instance=component_instance
        )

    def update_monitoring_status(self, status: Status) -> None:
        """Update the monitoring status.

        Args:
            status (Status): The status.

        """
        self._monit_status_service_subject.on_next(status)


__all__: list[str] = [
    "ComponentInstanceType",
    "ComponentTypeEnum",
    "MonitoredAbstract",
    "Status",
    "StatusService",
]



"""Provides security-related functions for the API."""

from asyncio import Task, TaskGroup
from http import HTTPStatus
from typing import Any, ClassVar, NewType, cast

import jwt
import pydantic
from fastapi import Request
from fastapi.exceptions import HTTPException
from pydantic import BaseModel

Scope = NewType("Scope", str)


class JWTBearerDecoded(BaseModel):
    """JWT bearer token decoded."""

    model_config: ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(
        arbitrary_types_allowed=True,
        extra="forbid",
        frozen=True,
    )

    scopes: list[str] | None = None


class JWTBearerAuthentication:
    """JWT Bearer Authentication.

    This class is used to authenticate users using JWT tokens.
    It extracts the token from the request, decodes it, and verifies its validity.
    """

    def __init__(self, scopes: list[Scope] | None = None, jwt_raw: str | None = None) -> None:
        """Initialize the OAuth2 class.

        Args:
            scopes (SecurityScopes): Security scopes for the OAuth2.
            jwt_raw (str): JWT token to be used for authentication.
        """
        self.jwt_raw: str | None = jwt_raw
        self.scopes: list[Scope] | None = scopes

    def _extract_raw_token(self, request: Request) -> str:
        """Extract the raw token from the request.

        Args:
            request (Request): FastAPI request object.

        Returns:
            str: Raw token.

        Raises:
            HTTPException: If the token is missing or invalid.
        """
        try:
            authorization_header: str | None = request.headers.get("Authorization")
        except (AttributeError, KeyError) as e:
            raise HTTPException(status_code=HTTPStatus.UNAUTHORIZED, detail="Missing Credentials") from e

        if not authorization_header:
            raise HTTPException(status_code=HTTPStatus.UNAUTHORIZED, detail="Missing Credentials")

        if not authorization_header.startswith("Bearer "):
            raise HTTPException(status_code=HTTPStatus.UNAUTHORIZED, detail="Invalid Credentials")

        return authorization_header.split(sep=" ")[1]

    async def _decode_jwt(self, jwt_raw: str) -> JWTBearerDecoded:
        """Decode the JWT token.

        Args:
            jwt_raw (str): Raw JWT token.

        Returns:
            JWTBearerDecoded: Decoded JWT token.

        Raises:
            HTTPException: If the token is invalid or expired.
        """
        try:
            jwt_decoded: dict[str, Any] = cast(
                dict[str, Any],
                jwt.decode(
                    jwt=jwt_raw,
                    algorithms=["HS256", "RS256"],
                    options={"verify_signature": True},
                ),
            )
            return JWTBearerDecoded(**jwt_decoded)
        except jwt.ExpiredSignatureError as e:
            raise HTTPException(status_code=HTTPStatus.UNAUTHORIZED, detail="Token expired") from e
        except jwt.InvalidTokenError as e:
            raise HTTPException(status_code=HTTPStatus.UNAUTHORIZED, detail="Invalid token") from e
        except pydantic.ValidationError as e:
            raise HTTPException(status_code=HTTPStatus.UNAUTHORIZED, detail=f"Invalid token: {e.json()}") from e

    async def _verify(self, jwt_raw: str) -> None:
        """Verify the JWT token.

        Args:
            jwt_raw (str): Raw JWT token.
        """
        pass

    def _has_scope(self, jwt_decoded: JWTBearerDecoded) -> None:
        """Check if the JWT token has the required scope.

        Args:
            jwt_decoded (JWTBearerDecoded): Decoded JWT token.

        """
        # Just Authentication (no scopes, no authorization)
        if not self.scopes:
            return
        # JWT without scopes (no authorization)
        if not jwt_decoded.scopes:
            raise HTTPException(status_code=HTTPStatus.FORBIDDEN, detail="Unauthorized")
        # Check if all required scopes are present
        if not all(scope in jwt_decoded.scopes for scope in (self.scopes or [])):
            raise HTTPException(status_code=HTTPStatus.FORBIDDEN, detail="Unauthorized")

        # All scopes are valid (authorization)
        return

    async def __call__(self, request: Request | None = None) -> JWTBearerDecoded:
        """Call the JWT bearer authentication.

        Args:
            request (Request): FastAPI request object.

        Returns:
            JWTBearerDecoded: Decoded JWT token.

        Raises:
            HTTPException: If the token is missing or invalid.
        """
        # Ensure that the jwt will be provided
        # by the request or by the jwt parameter
        if self.jwt_raw is None and request is None:
            raise HTTPException(status_code=HTTPStatus.UNAUTHORIZED, detail="Missing Credentials")
        jwt_raw: str
        if self.jwt_raw is None:
            jwt_raw = self._extract_raw_token(request=request)  # type: ignore[arg-type]
        else:
            jwt_raw = self.jwt_raw

        # Execute the io bound and cpu bound tasks in parallel
        async with TaskGroup() as tg:
            # TODO: Can be disabled by configuration (for operation purposes)
            # Ensure that the jwt is not revoked or expired
            tg.create_task(self._verify(jwt_raw=jwt_raw), name="verify_jwt")
            # Ensure that the jwt is not altered or expired
            task_decode: Task[Any] = tg.create_task(self._decode_jwt(jwt_raw=jwt_raw), name="decode_jwt")
        # Scope Validation
        jwt_decoded: JWTBearerDecoded = task_decode.result()
        self._has_scope(jwt_decoded=jwt_decoded)
        return jwt_decoded



"""ODM Plugin Module."""

from .depends import depends_odm_client, depends_odm_database
from .documents import BaseDocument
from .exceptions import (
    ODMPluginBaseException,
    ODMPluginConfigError,
    OperationError,
    UnableToCreateEntityDueToDuplicateKeyError,
)
from .helpers import PersistedEntity
from .plugins import ODMPlugin
from .repositories import AbstractRepository

__all__: list[str] = [
    "AbstractRepository",
    "BaseDocument",
    "ODMPlugin",
    "ODMPluginBaseException",
    "ODMPluginConfigError",
    "OperationError",
    "PersistedEntity",
    "UnableToCreateEntityDueToDuplicateKeyError",
    "depends_odm_client",
    "depends_odm_database",
]



"""Provides the Books API."""

from typing import cast
from uuid import UUID

from fastapi import APIRouter, Depends

from fastapi_factory_utilities.example.entities.books import BookEntity
from fastapi_factory_utilities.example.services.books import (
    BookService,
    depends_book_service,
)

from .responses import BookListReponse, BookResponseModel

api_v1_books_router: APIRouter = APIRouter(prefix="/books")
api_v2_books_router: APIRouter = APIRouter(prefix="/books")


@api_v1_books_router.get(path="", response_model=BookListReponse)
def get_books(
    books_service: BookService = Depends(depends_book_service),
) -> BookListReponse:
    """Get all books.

    Args:
        books_service (BookService): Book service.

    Returns:
        BookListReponse: List of books
    """
    books: list[BookEntity] = books_service.get_all_books()

    return BookListReponse(
        books=cast(
            list[BookResponseModel],
            map(lambda book: BookResponseModel(**book.model_dump()), books),
        ),
        size=len(books),
    )


@api_v1_books_router.get(path="/{book_id}", response_model=BookResponseModel)
def get_book(
    book_id: UUID,
    books_service: BookService = Depends(depends_book_service),
) -> BookResponseModel:
    """Get a book.

    Args:
        book_id (str): Book id
        books_service (BookService): Book service

    Returns:
        BookResponseModel: Book
    """
    book: BookEntity = books_service.get_book(book_id)

    return BookResponseModel(**book.model_dump())



"""Provides the abstract classes for the repositories."""

import datetime
from abc import ABC
from collections.abc import AsyncGenerator, Callable, Mapping
from contextlib import asynccontextmanager
from typing import Any, Generic, TypeVar, get_args
from uuid import UUID

from beanie import SortDirection
from motor.motor_asyncio import AsyncIOMotorClientSession, AsyncIOMotorDatabase
from pydantic import BaseModel
from pymongo.errors import DuplicateKeyError, PyMongoError
from pymongo.results import DeleteResult

from .documents import BaseDocument
from .exceptions import OperationError, UnableToCreateEntityDueToDuplicateKeyError

DocumentGenericType = TypeVar("DocumentGenericType", bound=BaseDocument)  # pylint: disable=invalid-name
EntityGenericType = TypeVar("EntityGenericType", bound=BaseModel)  # pylint: disable=invalid-name


def managed_session() -> Callable[[Callable[..., Any]], Callable[..., Any]]:
    """Decorator to manage the session.

    It will introspect the function arguments and check if the session is passed as a keyword argument.
    If it is not, it will create a new session and pass it to the function.
    """

    def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        async def wrapper(*args: Any, **kwargs: Any) -> Any:
            if "session" in kwargs:
                return await func(*args, **kwargs)

            async with args[0].get_session() as session:
                return await func(*args, **kwargs, session=session)

        return wrapper

    return decorator


class AbstractRepository(ABC, Generic[DocumentGenericType, EntityGenericType]):
    """Abstract class for the repository."""

    def __init__(self, database: AsyncIOMotorDatabase[Any]) -> None:
        """Initialize the repository."""
        super().__init__()
        self._database: AsyncIOMotorDatabase[Any] = database
        # Retrieve the generic concrete types
        generic_args: tuple[Any, ...] = get_args(self.__orig_bases__[0])  # type: ignore
        self._document_type: type[DocumentGenericType] = generic_args[0]
        self._entity_type: type[EntityGenericType] = generic_args[1]

    @asynccontextmanager
    async def get_session(self) -> AsyncGenerator[AsyncIOMotorClientSession, None]:
        """Yield a new session."""
        try:
            async with await self._database.client.start_session() as session:
                yield session
        except PyMongoError as error:
            raise OperationError(f"Failed to create session: {error}") from error

    @managed_session()
    async def insert(
        self, entity: EntityGenericType, session: AsyncIOMotorClientSession | None = None
    ) -> EntityGenericType:
        """Insert the entity into the database.

        Args:
            entity (EntityGenericType): The entity to insert.
            session (AsyncIOMotorClientSession | None): The session to use. Defaults to None. (managed by decorator)

        Returns:
            EntityGenericType: The entity created.

        Raises:
            ValueError: If the entity cannot be created from the document.
            UnableToCreateEntityDueToDuplicateKeyError: If the entity cannot be created due to a duplicate key error.
            OperationError: If the operation fails.
        """
        insert_time: datetime.datetime = datetime.datetime.now(tz=datetime.UTC)
        try:
            entity_dump: dict[str, Any] = entity.model_dump()
            entity_dump["created_at"] = insert_time
            entity_dump["updated_at"] = insert_time
            document: DocumentGenericType = self._document_type(**entity_dump)

        except ValueError as error:
            raise ValueError(f"Failed to create document from entity: {error}") from error

        try:
            document_created: DocumentGenericType = await document.insert(session=session)
        except DuplicateKeyError as error:
            raise UnableToCreateEntityDueToDuplicateKeyError(f"Failed to insert document: {error}") from error
        except PyMongoError as error:
            raise OperationError(f"Failed to insert document: {error}") from error

        try:
            entity_created: EntityGenericType = self._entity_type(**document_created.model_dump())
        except ValueError as error:
            raise ValueError(f"Failed to create entity from document: {error}") from error

        return entity_created

    @managed_session()
    async def update(
        self, entity: EntityGenericType, session: AsyncIOMotorClientSession | None = None
    ) -> EntityGenericType:
        """Update the entity in the database.

        Args:
            entity (EntityGenericType): The entity to update.
            session (AsyncIOMotorClientSession | None): The session to use. Defaults to None. (managed by decorator)

        Returns:
            EntityGenericType: The updated entity.

        Raises:
            ValueError: If the entity cannot be created from the document.
            OperationError: If the operation fails.
        """
        update_time: datetime.datetime = datetime.datetime.now(tz=datetime.UTC)
        try:
            entity_dump: dict[str, Any] = entity.model_dump()
            entity_dump["updated_at"] = update_time
            document: DocumentGenericType = self._document_type(**entity_dump)

        except ValueError as error:
            raise ValueError(f"Failed to create document from entity: {error}") from error

        try:
            document_updated: DocumentGenericType = await document.save(session=session)
        except PyMongoError as error:
            raise OperationError(f"Failed to update document: {error}") from error

        try:
            entity_updated: EntityGenericType = self._entity_type(**document_updated.model_dump())
        except ValueError as error:
            raise ValueError(f"Failed to create entity from document: {error}") from error

        return entity_updated

    @managed_session()
    async def get_one_by_id(
        self,
        entity_id: UUID,
        session: AsyncIOMotorClientSession | None = None,
    ) -> EntityGenericType | None:
        """Get the entity by its ID.

        Args:
            entity_id (UUID): The ID of the entity.
            session (AsyncIOMotorClientSession | None): The session to use. Defaults to None. (managed by decorator)

        Returns:
            EntityGenericType | None: The entity or None if not found.

        Raises:
            OperationError: If the operation fails.

        """
        try:
            document: DocumentGenericType | None = await self._document_type.get(document_id=entity_id, session=session)
        except PyMongoError as error:
            raise OperationError(f"Failed to get document: {error}") from error

        # If no document is found, return None
        if document is None:
            return None

        # Convert the document to an entity
        try:
            entity: EntityGenericType = self._entity_type(**document.model_dump())
        except ValueError as error:
            raise ValueError(f"Failed to create entity from document: {error}") from error

        return entity

    @managed_session()
    async def delete_one_by_id(
        self, entity_id: UUID, raise_if_not_found: bool = False, session: AsyncIOMotorClientSession | None = None
    ) -> None:
        """Delete a document by its ID.

        Args:
            entity_id (UUID): The ID of the entity.
            raise_if_not_found (bool, optional): Raise an exception if the document is not found. Defaults to False.
            session (AsyncIOMotorClientSession | None, optional): The session to use.
            Defaults to None. (managed by decorator)

        Raises:
            ValueError: If the document is not found and raise_if_not_found is True.
            OperationError: If the operation fails.

        """
        try:
            document_to_delete: DocumentGenericType | None = await self._document_type.get(
                document_id=entity_id, session=session
            )
        except PyMongoError as error:
            raise OperationError(f"Failed to get document to delete: {error}") from error

        if document_to_delete is None:
            if raise_if_not_found:
                raise ValueError(f"Failed to find document with ID {entity_id}")
            return

        try:
            delete_result: DeleteResult | None = await document_to_delete.delete()
        except PyMongoError as error:
            raise OperationError(f"Failed to delete document: {error}") from error

        if delete_result is not None and delete_result.deleted_count == 1 and delete_result.acknowledged:
            return

        raise OperationError("Failed to delete document.")

    @managed_session()
    async def find(  # noqa: PLR0913
        self,
        *args: Mapping[str, Any] | bool,
        projection_model: None = None,
        skip: int | None = None,
        limit: int | None = None,
        sort: None | str | list[tuple[str, SortDirection]] = None,
        session: AsyncIOMotorClientSession | None = None,
        ignore_cache: bool = False,
        fetch_links: bool = False,
        lazy_parse: bool = False,
        nesting_depth: int | None = None,
        nesting_depths_per_field: dict[str, int] | None = None,
        **pymongo_kwargs: Any,
    ) -> list[EntityGenericType]:
        """Find documents in the database.

        Args:
            *args: The arguments to pass to the find method.
            projection_model: The projection model to use.
            skip: The number of documents to skip.
            limit: The number of documents to return.
            sort: The sort order.
            session: The session to use.
            ignore_cache: Whether to ignore the cache.
            fetch_links: Whether to fetch links.
            lazy_parse: Whether to lazy parse the documents.
            nesting_depth: The nesting depth.
            nesting_depths_per_field: The nesting depths per field.
            **pymongo_kwargs: Additional keyword arguments to pass to the find method.

        Returns:
            list[EntityGenericType]: The list of entities.

        Raises:
            OperationError: If the operation fails.
            ValueError: If the entity cannot be created from the document.
        """
        try:
            documents: list[DocumentGenericType] = await self._document_type.find(
                *args,
                projection_model=projection_model,
                skip=skip,
                limit=limit,
                sort=sort,
                session=session,
                ignore_cache=ignore_cache,
                fetch_links=fetch_links,
                lazy_parse=lazy_parse,
                nesting_depth=nesting_depth,
                nesting_depths_per_field=nesting_depths_per_field,
                **pymongo_kwargs,
            ).to_list()
        except PyMongoError as error:
            raise OperationError(f"Failed to find documents: {error}") from error

        try:
            entities: list[EntityGenericType] = [self._entity_type(**document.model_dump()) for document in documents]
        except ValueError as error:
            raise ValueError(f"Failed to create entity from document: {error}") from error

        return entities



"""Provides the objects for the Hydra service."""

from typing import ClassVar

from pydantic import BaseModel, ConfigDict


class HydraTokenIntrospectObject(BaseModel):
    """Represents the object returned by the Hydra token introspection."""

    model_config: ClassVar[ConfigDict] = ConfigDict(extra="ignore")

    active: bool
    aud: list[str]
    client_id: str
    exp: int
    ext: dict[str, str] | None = None
    iat: int
    iss: str
    nbf: int
    obfuscated_subject: str | None = None
    scope: str
    sub: str
    token_type: str
    token_use: str
    username: str | None = None



"""Provides the abstract class for the publisher port for the Aiopika plugin."""

from typing import Any, ClassVar, Generic, Self, TypeVar

from aio_pika.abc import TimeoutType
from aio_pika.message import Message
from aiormq.abc import ConfirmationFrameType, DeliveredMessage
from pamqp.commands import Basic

from ..abstract import AbstractAiopikaResource
from ..exceptions import AiopikaPluginBaseError
from ..exchange import Exchange
from ..message import AbstractMessage

GenericMessage = TypeVar("GenericMessage", bound=AbstractMessage[Any])


class AbstractPublisher(AbstractAiopikaResource, Generic[GenericMessage]):
    """Abstract class for the publisher port for the Aiopika plugin."""

    DEFAULT_OPERATION_TIMEOUT: ClassVar[TimeoutType] = 10.0

    def __init__(self, exchange: Exchange, name: str | None = None) -> None:
        """Initialize the publisher port."""
        super().__init__()
        self._name: str = name or self.__class__.__name__
        self._exchange: Exchange = exchange

    async def setup(self) -> Self:
        """Setup the publisher."""
        await super().setup()
        await self._exchange.setup()
        return self

    async def publish(self, message: GenericMessage, routing_key: str) -> None:
        """Publish a message."""
        # Transform the message to an Aiopika message
        aiopika_message: Message
        try:
            aiopika_message = message.to_aiopika_message()
        except Exception as exception:
            raise AiopikaPluginBaseError(
                message="Failed to convert the message to an Aiopika message.",
            ) from exception
        # Publish the message
        confirmation: ConfirmationFrameType | DeliveredMessage | None
        try:
            confirmation = await self._exchange.exchange.publish(  # pyright: ignore
                message=aiopika_message,
                routing_key=routing_key,
                mandatory=True,
                timeout=self.DEFAULT_OPERATION_TIMEOUT,
            )
        except Exception as exception:
            raise AiopikaPluginBaseError(
                message="Failed to publish the message.",
            ) from exception

        if confirmation is None:
            raise AiopikaPluginBaseError(
                message="Failed to publish the message.",
            )
        if isinstance(confirmation, Basic.Return):
            raise AiopikaPluginBaseError(
                message="Failed to publish the message.",
            )



"""Provides a class for reading YAML files and converting them to Pydantic models."""

# mypy: disable-error-code="unused-ignore"

import os
import re
from pathlib import Path
from typing import Any, cast

from structlog.stdlib import BoundLogger, get_logger
from yaml import SafeLoader

logger: BoundLogger = get_logger()


class UnableToReadYamlFileError(Exception):
    """Raised when there is an error reading a YAML file."""

    def __init__(self, file_path: Path | None = None, message: str = "") -> None:
        """Initializes the exception.

        Args:
          file_path (str): The path to the YAML file.
          message (str): The error message.
        """
        super().__init__(f"Error reading YAML file: {file_path} - {message}")


class YamlFileReader:
    """Handles reading YAML files and converting them to Pydantic models."""

    re_pattern: re.Pattern[str] = re.compile(r"\${([A-Za-z0-9\-\_]+):?([A-Za-z0-9\-\_\/\:\.]*)?}")

    def __init__(
        self,
        file_path: Path,
        yaml_base_key: str | None = None,
        use_environment_injection: bool = True,
    ) -> None:
        """Initializes the YAML file reader.

        Args:
          file_path (str): The path to the YAML file.
          yaml_base_key (str | None, optional): The base key
          in the YAML file to read from. Defaults to None.
          use_environment_injection (bool, optional): Whether to use
          environment injection. Defaults to True.
        """
        # Store the file path and base key for YAML reading
        self._yaml_base_key: str | None = yaml_base_key
        self._file_path: Path = file_path

        # Store whether to use environment injection
        self._use_environment_injection: bool = use_environment_injection

    def _filter_data_with_base_key(self, yaml_data: dict[str, Any]) -> dict[str, Any] | None:
        """Extracts the data from the YAML file with the base key.

        Args:
            yaml_data (dict): The data from the YAML file.

        Returns:
            dict: The filtered data from the YAML file.

        Raises:
            KeyError: If the base key is not found in the YAML file.
        """
        if self._yaml_base_key is not None:
            keys: list[str] = self._yaml_base_key.split(".")
            while len(keys) != 0:
                key: str = keys.pop(0)
                try:
                    yaml_data = yaml_data[key]
                except KeyError:
                    logger.warning(f"Base key {key} not found in YAML file" + " from {self._yaml_base_key}")
                    return dict()
        return yaml_data

    def _read_yaml_file(self, file_path: Path) -> dict[str, Any]:
        """Reads the YAML file and returns the data as a dictionary.

        Args:
            file_path (Path): The path to the YAML file.

        Returns:
            dict: The data from the YAML file.

        Raises:
            ValueError: If there is an error reading the file.
            FileNotFoundError: If the file is not found.
        """
        if not os.path.exists(file_path):
            raise FileNotFoundError(f"File not found: {file_path}")

        with open(file=file_path, encoding="UTF-8") as file:
            loader = SafeLoader(file)

            try:
                yaml_data: dict[str, Any] = cast(
                    dict[str, Any],
                    loader.get_data(),  # type: ignore
                )
            except Exception as exception:
                raise ValueError(f"Error reading YAML file: {file_path}") from exception

            return yaml_data

    def _inject_environment_variables(
        self, yaml_data: dict[str, Any] | str | list[str] | bool | int
    ) -> dict[str, Any] | str | list[str] | bool | int:
        """Injects environment variables into the YAML data recursively.

        Args:
            yaml_data (dict | str | list | bool): The data from the YAML file.

        Returns:
            dict: The data from the YAML file
            with environment variables injected.

        Raises:
            ValueError: If the YAML data is None.
        """
        if isinstance(yaml_data, dict):
            for key, value in yaml_data.items():
                yaml_data[key] = self._inject_environment_variables(value)
        elif isinstance(yaml_data, list):
            yaml_data = [cast(str, self._inject_environment_variables(yaml_data=value)) for value in yaml_data]
        elif isinstance(yaml_data, bool) or isinstance(yaml_data, int):
            return yaml_data
        elif isinstance(yaml_data, str):  # type: ignore
            while True:
                match = self.re_pattern.search(yaml_data)
                if match is None:
                    break
                env_key = match.group(1)
                env_default = match.group(2)
                env_value = os.getenv(env_key, env_default)
                yaml_data = yaml_data.replace(match.group(0), env_value)
        else:
            raise ValueError(f"Type not supported: {type(yaml_data)}")
        return yaml_data

    def read(self) -> dict[str, Any]:
        """Reads the YAML file and converts it to a Pydantic model with env injected.

        Raises:
            UnableToReadYamlFileError: If there is an error reading the file.
        """
        # Read the YAML file and filter the data with the base key
        try:
            yaml_data: dict[str, Any] | None = self._filter_data_with_base_key(
                self._read_yaml_file(file_path=self._file_path)
            )
        except (FileNotFoundError, ValueError, KeyError) as exception:
            raise UnableToReadYamlFileError(file_path=self._file_path, message=str(exception)) from exception

        if yaml_data is None:
            return dict()

        if self._use_environment_injection:
            yaml_data_with_env_injected: dict[str, Any] = cast(
                dict[str, Any], self._inject_environment_variables(yaml_data)
            )
            return dict[str, Any](yaml_data_with_env_injected)
        else:
            return dict[str, Any](yaml_data)



"""Readiness calculator strategies."""

from typing import Protocol

from .enums import ReadinessStatusEnum
from .types import ComponentInstanceKey, Status


class ReadinessCalculatorStrategy(Protocol):
    """Readiness calculator strategy."""

    def __init__(self, components_status: dict[ComponentInstanceKey, Status]) -> None:
        """Initialize the readiness calculator.

        Args:
            components_status (dict[str, Status]): The components status.
        """
        raise NotImplementedError

    def calculate(self) -> ReadinessStatusEnum:
        """Calculate the readiness status."""
        raise NotImplementedError


class ReadinessSimpleCalculatorStrategy:
    """Readiness calculator."""

    def __init__(self, components_status: dict[ComponentInstanceKey, Status]) -> None:
        """Initialize the readiness calculator.

        Args:
            components_status (dict[str, Status]): The components status.
        """
        self._components_status: dict[ComponentInstanceKey, Status] = components_status

    def calculate(self) -> ReadinessStatusEnum:
        """Calculate the readiness status."""
        for status in self._components_status.values():
            if status["readiness"] != ReadinessStatusEnum.READY:
                return ReadinessStatusEnum.NOT_READY
        return ReadinessStatusEnum.READY



"""Provides the module for the ODM plugin."""

from typing import Any, ClassVar, Self

from bson import CodecOptions
from motor.motor_asyncio import AsyncIOMotorClient, AsyncIOMotorDatabase
from pymongo.server_api import ServerApi, ServerApiVersion
from structlog.stdlib import get_logger

from fastapi_factory_utilities.core.protocols import ApplicationAbstractProtocol
from fastapi_factory_utilities.core.utils.importlib import get_path_file_in_package
from fastapi_factory_utilities.core.utils.yaml_reader import (
    UnableToReadYamlFileError,
    YamlFileReader,
)

from .configs import ODMConfig
from .exceptions import ODMPluginConfigError

_logger = get_logger()


class ODMBuilder:
    """Factory to create the resources for the ODM plugin.

    The factory is responsible for creating the resources for the ODM plugin.
    - The ODM configuration.
    - The ODM client.
    - The ODM database.

    ```python
    # Example of using the ODMFactory
    odm_factory: ODMFactory = ODMFactory(application=application)
    odm_factory.build_all()
    # Access the ODM database created
    database: AsyncIOMotorDatabase[Any] = odm_factory.database
    ```

    """

    MS_TO_S: ClassVar[int] = 1000
    SLEEP_TIME_S: ClassVar[float] = 0.001

    def __init__(
        self,
        application: ApplicationAbstractProtocol,
        odm_config: ODMConfig | None = None,
        odm_client: AsyncIOMotorClient[Any] | None = None,
        odm_database: AsyncIOMotorDatabase[Any] | None = None,
    ) -> None:
        """Initialize the ODMFactory.

        Args:
            application (BaseApplicationProtocol): The application.
            odm_config (ODMConfig): The ODM configuration for injection. (Default is None)
            odm_client (AsyncIOMotorClient): The ODM client for injection. (Default is None)
            odm_database (AsyncIOMotorDatabase): The ODM database for injection. (Default is None)

        """
        self._application: ApplicationAbstractProtocol = application
        self._config: ODMConfig | None = odm_config
        self._odm_client: AsyncIOMotorClient[Any] | None = odm_client
        self._odm_database: AsyncIOMotorDatabase[Any] | None = odm_database

    @property
    def config(self) -> ODMConfig | None:
        """Provide the ODM configuration object.

        Returns:
            ODMConfig: The ODM configuration object.
        """
        return self._config

    @property
    def odm_client(self) -> AsyncIOMotorClient[Any] | None:
        """Provide the ODM client.

        Returns:
            AsyncIOMotorClient | None: The ODM client.
        """
        return self._odm_client

    @property
    def odm_database(self) -> AsyncIOMotorDatabase[Any] | None:
        """Provide the ODM database.

        Returns:
            AsyncIOMotorDatabase | None: The ODM database.
        """
        return self._odm_database

    def build_odm_config(
        self,
    ) -> Self:
        """Build the ODM configuration object.

        Returns:
            Self: The ODM factory.

        Raises:
            ODMPluginConfigError: If the package name is not set or the configuration file is not found.
        """
        if self._config is not None:
            return self

        if self._application.PACKAGE_NAME == "":
            raise ODMPluginConfigError("The package name must be set in the concrete application class.")
        # Read the application configuration file
        try:
            yaml_file_content: dict[str, Any] = YamlFileReader(
                file_path=get_path_file_in_package(
                    filename="application.yaml",
                    package=self._application.PACKAGE_NAME,
                ),
                yaml_base_key="odm",
                use_environment_injection=True,
            ).read()
        except (FileNotFoundError, ImportError, UnableToReadYamlFileError) as exception:
            raise ODMPluginConfigError("Unable to read the application configuration file.") from exception

        # Create the application configuration model
        try:
            self._config = ODMConfig(**yaml_file_content)
        except ValueError as exception:
            raise ODMPluginConfigError("Unable to create the application configuration model.") from exception
        return self

    # ======
    # KEEP IT, Waiting for additional tests
    # @classmethod
    # def _wait_client_to_be_ready(cls, client: AsyncIOMotorClient[Any], timeout_s: int) -> None:
    #     """Wait for the ODM client to be ready.

    #     Args:
    #         client (AsyncIOMotorClient): The ODM client.
    #         timeout_s (int): The timeout in seconds.

    #     Raises:
    #         TimeoutError: If the ODM client is not ready in the given timeout.
    #     """
    #     start_time: float = time.time()
    #     message_time: float = time.time()
    #     while (time.time() - start_time) < (timeout_s):
    #         if len(client.nodes) > 0:  # type: ignore
    #             _logger.info(f"Waiting {(time.time() - start_time)*cls.MS_TO_S}ms for the ODM client to be ready.")
    #             return

    #         if (time.time() - message_time) > 1:
    #             elaps_time: float = time.time() - start_time
    #             _logger.debug(f"Waiting for the ODM client to be ready. (from {int(elaps_time)}s) ")
    #             message_time = time.time()
    #         time.sleep(cls.SLEEP_TIME_S)

    #     raise TimeoutError("The ODM client is not ready in the given timeout.")
    # ======

    def build_client(
        self,
    ) -> Self:
        """Build the ODM client.

        Returns:
            Self: The ODM factory.

        Raises:
            ODMPluginConfigError: If the ODM configuration is not build or provided.
        """
        if self._odm_client is not None:
            return self

        if self._config is None:
            raise ODMPluginConfigError(
                "ODM configuration is not set. Provide the ODM configuration using "
                "build_odm_config method or through parameter."
            )

        self._odm_client = AsyncIOMotorClient(
            host=self._config.uri,
            connect=True,
            connectTimeoutMS=self._config.connection_timeout_ms,
            serverSelectionTimeoutMS=self._config.connection_timeout_ms,
            server_api=ServerApi(version=ServerApiVersion.V1),
            tz_aware=True,
        )

        # KEEP IT, Waiting for additional tests
        # self._wait_client_to_be_ready(client=self._odm_client, timeout_s=self._config.connection_timeout_s)

        return self

    def build_database(
        self,
    ) -> Self:
        """Build the ODM database.

        The ODM client and ODM configuration are recommended to be provided through call to the build_client and
        build_odm_config methods.

        Returns:
            Any: The ODM database.

        Raises:
            ODMPluginConfigError: If the ODM configuration is not build or provided.
        """
        if self._odm_database is not None:
            return self

        if self._config is None:
            raise ODMPluginConfigError(
                "ODM configuration is not set. Provide the ODM configuration using "
                "build_odm_config method or through parameter."
            )

        database_name: str = self._config.database

        if self._odm_client is None:
            raise ODMPluginConfigError(
                "ODM client is not set. Provide the ODM client using build_client method or through parameter."
            )

        self._odm_database = self._odm_client.get_database(
            name=database_name,
            codec_options=CodecOptions(  # pyright: ignore[reportUnknownArgumentType]
                tz_aware=True,
            ),
        )

        return self

    def build_all(self) -> Self:
        """Build all the resources for the ODM plugin.

        Returns:
            Self: The ODM factory.

        Raises:
            ODMPluginConfigError: If the ODM configuration is not build or provided.
        """
        self.build_odm_config()
        self.build_client()
        self.build_database()

        return self

    async def wait_ping(self) -> None:
        """Wait for the ODM client to be ready.

        Returns:
            Self: The ODM factory.
        """
        if self._odm_client is None:
            raise ODMPluginConfigError(
                "ODM client is not set. Provide the ODM client using build_client method or through parameter."
            )

        try:
            await self._odm_client.admin.command("ping")
        except Exception as exception:  # pylint: disable=broad-except
            raise ODMPluginConfigError("Unable to ping the ODM client.") from exception



"""Provides a service to interact with the Hydra service."""

from base64 import b64encode
from http import HTTPStatus
from typing import Annotated

import aiohttp
from fastapi import Depends
from pydantic import ValidationError

from fastapi_factory_utilities.core.app import (
    DependencyConfig,
    HttpServiceDependencyConfig,
    depends_dependency_config,
)

from .exceptions import HydraOperationError, HydraTokenInvalidError
from .objects import HydraTokenIntrospectObject


class HydraService:
    """Service to interact with the Hydra service."""

    INTROSPECT_ENDPOINT: str = "/admin/oauth2/introspect"
    CLIENT_CREDENTIALS_ENDPOINT: str = "/oauth2/token"

    def __init__(
        self,
        hydra_admin_http_config: HttpServiceDependencyConfig,
        hydra_public_http_config: HttpServiceDependencyConfig,
    ) -> None:
        """Instanciate the Hydra service.

        Args:
            hydra_admin_http_config (HttpServiceDependencyConfig): The Hydra admin HTTP configuration.
            hydra_public_http_config (HttpServiceDependencyConfig): The Hydra public HTTP configuration.
        """
        self._hydra_admin_http_config: HttpServiceDependencyConfig = hydra_admin_http_config
        self._hydra_public_http_config: HttpServiceDependencyConfig = hydra_public_http_config

    async def introspect(self, token: str) -> HydraTokenIntrospectObject:
        """Introspects a token using the Hydra service.

        Args:
            token (str): The token to introspect.

        Raises:
            HydraOperationError: If the introspection fails.
            HydraTokenInvalidError: If the token is invalid.
        """
        async with aiohttp.ClientSession(
            base_url=str(self._hydra_admin_http_config.url),
        ) as session:
            async with session.post(
                url=self.INTROSPECT_ENDPOINT,
                data={"token": token},
            ) as response:
                if response.status != HTTPStatus.OK:
                    raise HydraTokenInvalidError()

                try:
                    instrospect: HydraTokenIntrospectObject = HydraTokenIntrospectObject(**await response.json())
                except ValidationError as error:
                    raise HydraOperationError() from error

                return instrospect

    async def oauth2_client_credentials(self, client_id: str, client_secret: str, scope: str) -> str:
        """Get the OAuth2 client credentials.

        Args:
            client_id (str): The client ID.
            client_secret (str): The client secret.
            scope (str): The scope.

        Returns:
            str: The access token.

        Raises:
            HydraOperationError: If the client credentials request fails.
        """
        # Create base64 encoded Basic Auth header
        auth_string = f"{client_id}:{client_secret}"
        auth_bytes = auth_string.encode("utf-8")
        auth_b64 = b64encode(auth_bytes).decode("utf-8")

        async with aiohttp.ClientSession(
            base_url=str(self._hydra_public_http_config.url),
        ) as session:
            async with session.post(
                url=self.CLIENT_CREDENTIALS_ENDPOINT,
                headers={"Authorization": f"Basic {auth_b64}"},
                data={"grant_type": "client_credentials", "scope": scope},
            ) as response:
                response_data = await response.json()
                if response.status != HTTPStatus.OK:
                    raise HydraOperationError(f"Failed to get client credentials: {response_data}")

                return response_data["access_token"]


def depends_hydra_service(
    dependency_config: Annotated[DependencyConfig, Depends(depends_dependency_config)],
) -> HydraService:
    """Dependency injection for the Hydra service.

    Args:
        dependency_config (DependencyConfig): The dependency configuration.

    Returns:
        HydraService: The Hydra service instance.

    Raises:
        HydraOperationError: If the Hydra admin or public dependency is not configured.
    """
    if dependency_config.hydra_admin is None or dependency_config.hydra_public is None:
        raise HydraOperationError(message="Hydra admin or public dependency not configured")

    return HydraService(
        hydra_admin_http_config=dependency_config.hydra_admin,
        hydra_public_http_config=dependency_config.hydra_public,
    )
